import unittest
import os

import numpy as np
import numpy.testing as npt
import openmdao.api as om

import wisdem.rotorse.rotor_elasticity as rel
from wisdem.commonse.cross_sections import Tube
from wisdem.rotorse.rotor_elasticity import generate_KI, KI_to_Elastic
import wisdem.precomp.properties as prop
import wisdem.precomp.precomp_to_beamdyn as pc2bd

try:
   import cPickle as pickle
except Exception:
   import pickle

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
# https://stackoverflow.com/questions/20716812/saving-and-loading-multiple-objects-in-pickle-file
def loadall(filename):
    with open(os.path.join(mydir, filename), "rb") as f:
        while True:
            try:
                yield pickle.load(f)
            except EOFError:
                break

class TestRE(unittest.TestCase):
    def setUp(self):
        self.discrete_inputs = {}
        self.discrete_outputs = {}
        self.inputs = {}
        self.outputs = {}
        self.options = {}
        self.options = {}
        self.options["WISDEM"] = {}
        self.options["WISDEM"]["RotorSE"] = {}
        self.options["WISDEM"]["RotorSE"]["n_xy"] = npts = 101
        self.options["WISDEM"]["RotorSE"]["n_span"] = nspan = 31
        self.options["WISDEM"]["RotorSE"]["n_webs"] = nweb = 0
        self.options["WISDEM"]["RotorSE"]["n_layers"] = nlay = 1
        self.options["WISDEM"]["RotorSE"]["te_ss"] = "none"
        self.options["WISDEM"]["RotorSE"]["te_ps"] = "none"
        self.options["WISDEM"]["RotorSE"]["spar_cap_ss"] = "none"
        self.options["WISDEM"]["RotorSE"]["spar_cap_ps"] = "none"
        self.options["WISDEM"]["RotorSE"]["layer_name"] = ["mylayer"]
        self.options["WISDEM"]["RotorSE"]["layer_mat"] = ["mymat"]
        self.options["materials"] = {}
        self.options["materials"]["n_mat"] = nmat = 1
        self.options["General"] = {}
        self.options["General"]["verbosity"] = False

        # Use a tubular blade
        self.inputs["r"] = np.linspace(0.0, 10.0, nspan)
        self.inputs["theta"] = np.zeros(nspan)
        self.inputs["chord"] = 2 * np.ones(nspan)
        self.inputs["pitch_axis"] = 0.5 * np.ones(nspan) # 0=LE, 1=TE
        self.inputs["precurve"] = np.zeros(nspan)
        self.inputs["presweep"] = np.zeros(nspan)
        self.inputs["coord_xy_interp"] = np.zeros( (nspan, npts, 2) )
        angle = np.linspace(0, 2*np.pi, npts)
        for k in range(nspan):
            self.inputs["coord_xy_interp"][k,:,0] = 0.5 * np.cos(angle) + 0.5
            self.inputs["coord_xy_interp"][k,:,1] = 0.5 * np.sin(angle)
        self.inputs["coord_xy_interp"][:,0,1] = 0.0
        self.inputs["coord_xy_interp"][:,-1,1] = 0.0
        self.inputs["uptilt"] = np.zeros(1)
        self.discrete_inputs["n_blades"] = 3
        self.inputs["web_start_nd"] = self.inputs["web_end_nd"] = np.zeros((nweb, nspan))
        self.inputs["layer_web"] = np.zeros(nlay)
        self.discrete_inputs["definition_layer"] = np.ones(nlay)
        self.inputs["layer_thickness"] = 0.01 * np.ones((nlay, nspan))
        self.inputs["layer_start_nd"] = np.zeros((nlay, nspan))
        self.inputs["layer_end_nd"] = np.ones((nlay, nspan))
        self.inputs["fiber_orientation"] = 90 * np.ones((nlay, nspan))

        self.discrete_inputs["mat_name"] = ["mymat"]
        self.discrete_inputs["orth"] = np.zeros(nmat)
        self.inputs["E"] = 1e10 * np.ones((nmat,3))
        self.inputs["G"] = 1e8 * np.ones((nmat,3))
        self.inputs["nu"] = 0.33 * np.ones((nmat,3))
        self.inputs["rho"] = 1e3 * np.ones(nmat)
        self.inputs["joint_position"] = self.inputs["joint_mass"] = 0.0

        self.mytube = Tube(self.inputs["chord"][0], self.inputs["layer_thickness"][0,0])

    def run_precomp(self):

        self.myobj = rel.RunPreComp(modeling_options=self.options, opt_options={})
        rotorse_options = self.options["WISDEM"]["RotorSE"]
        self.myobj.n_span = rotorse_options["n_span"]
        self.myobj.n_webs = rotorse_options["n_webs"]
        self.myobj.n_layers = rotorse_options["n_layers"]
        self.myobj.n_xy = rotorse_options["n_xy"]  # Number of coordinate points to describe the airfoil geometry
        mat_init_options = self.options["materials"]
        self.myobj.n_mat = mat_init_options["n_mat"]
        self.myobj.verbosity = self.options["General"]["verbosity"]

        self.myobj.te_ss_var = rotorse_options["te_ss"]
        self.myobj.te_ps_var = rotorse_options["te_ps"]
        self.myobj.spar_cap_ss_var = rotorse_options["spar_cap_ss"]
        self.myobj.spar_cap_ps_var = rotorse_options["spar_cap_ps"]
        self.myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        for k in self.outputs.keys():
            if type(self.outputs[k]) == type(np.array([])):
                self.inputs[k] = self.outputs[k]
            else:
                self.inputs[k] = np.array([ self.outputs[k] ])
        self.myobj = rel.TotalBladeProperties(modeling_options=self.options, opt_options={})

    def test_no_pitch(self):
        self.run_precomp()
        self.myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["z"], self.inputs["r"])
        npt.assert_almost_equal(self.outputs["x_cg"], 0.0)
        npt.assert_almost_equal(self.outputs["x_tc"], 0.0)
        npt.assert_almost_equal(self.outputs["x_sc"], 0.0)
        npt.assert_almost_equal(self.outputs["y_cg"], 0.0)
        npt.assert_almost_equal(self.outputs["y_tc"], 0.0)
        npt.assert_almost_equal(self.outputs["y_sc"], 0.0)
        npt.assert_almost_equal(self.outputs["EIxy"], 0.0, decimal=6)
        npt.assert_almost_equal(self.outputs["A"], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["EIxx"]/self.inputs["E"][0,0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(self.outputs["EIyy"]/self.inputs["E"][0,0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(self.outputs["GJ"]/self.inputs["G"][0,0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(self.outputs["EA"]/self.inputs["E"][0,0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["rhoA"]/self.inputs["rho"][0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["rhoJ"]/self.inputs["rho"][0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(self.outputs["Tw_iner"], 0.0)
        npt.assert_almost_equal(self.outputs["flap_iner"]/self.inputs["rho"][0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(self.outputs["edge_iner"]/self.inputs["rho"][0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(self.outputs["blade_mass"], self.mytube.Area*self.inputs["rho"][0]*self.inputs["r"][-1], decimal=-1)
        npt.assert_almost_equal(self.outputs["mass_all_blades"], 3*self.outputs["blade_mass"], decimal=3)
        idx = np.int_(np.floor(0.5*self.inputs["r"].size))
        npt.assert_almost_equal(self.outputs["blade_span_cg"], self.inputs["r"][idx], decimal=1)
        npt.assert_almost_equal(self.outputs["blade_moment_of_inertia"], self.outputs["blade_mass"]*self.inputs["r"][-1]**2/3.0, decimal=-1)

        '''
        self.outputs["xu_spar"]
        self.outputs["xl_spar"]
        self.outputs["yu_spar"]
        self.outputs["yl_spar"]
        self.outputs["xu_te"]
        self.outputs["xl_te"]
        self.outputs["yu_te"]
        self.outputs["yl_te"]
        self.outputs["I_all_blades"]
        '''
        
    def test_with_pitch(self):
        self.inputs["theta"] = 45 * np.ones(self.inputs["theta"].shape)
        
        self.run_precomp()
        self.myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["z"], self.inputs["r"])
        npt.assert_almost_equal(self.outputs["x_cg"], 0.0)
        npt.assert_almost_equal(self.outputs["x_tc"], 0.0)
        npt.assert_almost_equal(self.outputs["x_sc"], 0.0)
        npt.assert_almost_equal(self.outputs["y_cg"], 0.0)
        npt.assert_almost_equal(self.outputs["y_tc"], 0.0)
        npt.assert_almost_equal(self.outputs["y_sc"], 0.0)
        npt.assert_almost_equal(self.outputs["EIxy"], 0.0, decimal=6)
        npt.assert_almost_equal(self.outputs["A"], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["EIxx"]/self.inputs["E"][0,0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(self.outputs["EIyy"]/self.inputs["E"][0,0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(self.outputs["GJ"]/self.inputs["G"][0,0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(self.outputs["EA"]/self.inputs["E"][0,0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["rhoA"]/self.inputs["rho"][0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["rhoJ"]/self.inputs["rho"][0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(self.outputs["Tw_iner"], 45.0)
        npt.assert_almost_equal(self.outputs["flap_iner"]/self.inputs["rho"][0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(self.outputs["edge_iner"]/self.inputs["rho"][0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(self.outputs["blade_mass"], self.mytube.Area*self.inputs["rho"][0]*self.inputs["r"][-1], decimal=-1)
        npt.assert_almost_equal(self.outputs["mass_all_blades"], 3*self.outputs["blade_mass"], decimal=3)
        idx = np.int_(np.floor(0.5*self.inputs["r"].size))
        npt.assert_almost_equal(self.outputs["blade_span_cg"], self.inputs["r"][idx], decimal=1)
        npt.assert_almost_equal(self.outputs["blade_moment_of_inertia"], self.outputs["blade_mass"]*self.inputs["r"][-1]**2/3.0, decimal=-1)
        
        
    def test_with_le_pitch_axis(self):
        self.inputs["pitch_axis"] = np.zeros(self.inputs["pitch_axis"].shape)
        
        self.run_precomp()
        self.myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(self.outputs["z"], self.inputs["r"])
        npt.assert_almost_equal(self.outputs["x_cg"], 0.0)
        npt.assert_almost_equal(self.outputs["x_tc"], 0.0)
        npt.assert_almost_equal(self.outputs["x_sc"], 0.0)
        npt.assert_almost_equal(self.outputs["y_cg"], 1.0)
        npt.assert_almost_equal(self.outputs["y_tc"], 1.0)
        npt.assert_almost_equal(self.outputs["y_sc"], 1.0)
        npt.assert_almost_equal(self.outputs["EIxy"], 0.0, decimal=6)
        npt.assert_almost_equal(self.outputs["A"], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["EIxx"]/self.inputs["E"][0,0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(self.outputs["EIyy"]/self.inputs["E"][0,0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(self.outputs["GJ"]/self.inputs["G"][0,0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(self.outputs["EA"]/self.inputs["E"][0,0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["rhoA"]/self.inputs["rho"][0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(self.outputs["rhoJ"]/self.inputs["rho"][0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(self.outputs["Tw_iner"], 0.0)
        npt.assert_almost_equal(self.outputs["flap_iner"]/self.inputs["rho"][0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(self.outputs["edge_iner"]/self.inputs["rho"][0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(self.outputs["blade_mass"], self.mytube.Area*self.inputs["rho"][0]*self.inputs["r"][-1], decimal=-1)
        npt.assert_almost_equal(self.outputs["mass_all_blades"], 3*self.outputs["blade_mass"], decimal=3)
        idx = np.int_(np.floor(0.5*self.inputs["r"].size))
        npt.assert_almost_equal(self.outputs["blade_span_cg"], self.inputs["r"][idx], decimal=1)
        npt.assert_almost_equal(self.outputs["blade_moment_of_inertia"], self.outputs["blade_mass"]*self.inputs["r"][-1]**2/3.0, decimal=-1)

    def test_KI_to_Elastic(self):

        fnames = ['../test_precomp/section_dump_iea15mw.pkl']
        
        for f in fnames:
            with self.subTest(f=f):
                myitems = loadall(f)
                nsec = myitems.__next__()
                for k in range(nsec):
                    with self.subTest(i=k):
                        chord = myitems.__next__()
                        theta = myitems.__next__()
                        th_prime = myitems.__next__()
                        le_loc = myitems.__next__()
                        xnode = myitems.__next__()
                        ynode = myitems.__next__()
                        E1 = myitems.__next__()
                        E2 = myitems.__next__()
                        G12 = myitems.__next__()
                        nu12 = myitems.__next__()
                        rho = myitems.__next__()
                        locU = myitems.__next__()
                        n_laminaU = myitems.__next__()
                        n_pliesU = myitems.__next__()
                        tU = myitems.__next__()
                        thetaU = myitems.__next__()
                        mat_idxU = myitems.__next__()
                        locL = myitems.__next__()
                        n_laminaL = myitems.__next__()
                        n_pliesL = myitems.__next__()
                        tL = myitems.__next__()
                        thetaL = myitems.__next__()
                        mat_idxL = myitems.__next__()
                        nwebs = myitems.__next__()
                        locW = myitems.__next__()
                        n_laminaW = myitems.__next__()
                        n_pliesW = myitems.__next__()
                        tW = myitems.__next__()
                        thetaW = myitems.__next__()
                        mat_idxW = myitems.__next__()

                        results_fort = myitems.__next__()
                        results_py = prop.properties(chord,
                                                     theta,
                                                     th_prime,
                                                     le_loc,
                                                     xnode,
                                                     ynode,
                                                     E1,
                                                     E2,
                                                     G12,
                                                     nu12,
                                                     rho,
                                                     locU,
                                                     n_laminaU,
                                                     n_pliesU,
                                                     tU,
                                                     thetaU,
                                                     mat_idxU,
                                                     locL,
                                                     n_laminaL,
                                                     n_pliesL,
                                                     tL,
                                                     thetaL,
                                                     mat_idxL,
                                                     nwebs,
                                                     locW,
                                                     n_laminaW,
                                                     n_pliesW,
                                                     tW,
                                                     thetaW,
                                                     mat_idxW,
                                                     )

                    if k==15:
                        (eifbar,eilbar,gjbar,eabar,eiflbar,
                        sfbar,slbar,sftbar,sltbar,satbar,
                        z_sc,y_sc,ztc_ref,ytc_ref,
                        mass,area,iflap_eta,ilag_zeta,tw_iner,
                        zcm_ref,ycm_ref) = results_py

                        EIxx  = eilbar
                        EIyy  = eifbar
                        GJ  = gjbar
                        EA  = eabar
                        EIxy  = eiflbar
                        EA_EIxx  = slbar
                        EA_EIyy  = sfbar
                        EIxx_GJ  = sltbar
                        EIyy_GJ  = sftbar
                        EA_GJ  = satbar
                        x_sc  = z_sc
                        y_sc  = y_sc
                        x_tc  = ztc_ref
                        y_tc  = ytc_ref
                        rhoA  = mass
                        A  = area
                        flap_iner  = iflap_eta
                        edge_iner  = ilag_zeta
                        Tw_iner  = tw_iner
                        x_cg  = zcm_ref
                        y_cg  = ycm_ref


                        # Build stiffness matrix at the reference axis
                        K_precomp = pc2bd.pc2bd_K(
                            EA,
                            EIxx,
                            EIyy,
                            EIxy,
                            EA_EIxx,
                            EA_EIyy,
                            EIxx_GJ,
                            EIyy_GJ,
                            EA_GJ,
                            GJ,
                            flap_iner+edge_iner,
                            edge_iner,
                            flap_iner,
                            x_sc,
                            y_sc,
                            )
                        # Build inertia matrix at the reference axis
                        I_precomp = pc2bd.pc2bd_I(
                            rhoA,
                            edge_iner,
                            flap_iner,
                            edge_iner+flap_iner,
                            x_cg,
                            y_cg,
                            np.deg2rad(Tw_iner),
                            np.deg2rad(theta),
                            )
                        

                        dummy_options = {}
                        dummy_options['WISDEM'] = {}
                        dummy_options['WISDEM']['RotorSE'] = {}
                        dummy_options['WISDEM']['RotorSE']["n_span"] = 1

                        prob = om.Problem()
                        prob.model.add_subsystem("KI_reverse", KI_to_Elastic(modeling_options=dummy_options), promotes=["*"])
                        prob.setup()

                        prob.set_val("K", K_precomp)
                        prob.set_val("I", I_precomp)
                        prob.set_val("theta", theta)

                        prob.run_model()
                        Tw_iner_back = prob.get_val("Tw_iner")
                        x_cg_back = prob.get_val("x_cg")
                        y_cg_back = prob.get_val("y_cg")
                        edge_iner_back = prob.get_val("edge_iner")
                        flap_iner_back = prob.get_val("flap_iner")
                        EIxx_back = prob.get_val("EIxx")
                        EIyy_back = prob.get_val("EIyy")
                        x_tc_back = prob.get_val("x_tc")
                        y_tc_back = prob.get_val("y_tc")
                        x_sc_back = prob.get_val("x_sc")
                        y_sc_back = prob.get_val("y_sc")
                        GJ_back = prob.get_val("GJ")
                        npt.assert_almost_equal(Tw_iner, Tw_iner_back, decimal=6)
                        npt.assert_almost_equal(x_cg, x_cg_back, decimal=6)
                        npt.assert_almost_equal(np.abs(edge_iner-edge_iner_back)/edge_iner, 0.0, decimal=6)
                        npt.assert_almost_equal(np.abs(flap_iner-flap_iner_back)/flap_iner, 0.0, decimal=6)
                        npt.assert_almost_equal(y_cg, y_cg_back, decimal=6)
                        npt.assert_almost_equal(x_tc, x_tc_back, decimal=2)
                        npt.assert_almost_equal(y_tc, y_tc_back, decimal=2)
                        npt.assert_almost_equal(x_sc, x_sc_back, decimal=2)
                        npt.assert_almost_equal(y_sc, y_sc_back, decimal=2)
                        npt.assert_almost_equal(np.abs(EIxx-EIxx_back)/EIxx, 0.0, decimal=6)
                        npt.assert_almost_equal(np.abs(EIyy-EIyy_back)/EIyy, 0.0, decimal=6)
                        npt.assert_almost_equal(np.abs(GJ-GJ_back)/GJ, 0.0, decimal=6)

        
if __name__ == "__main__":
    unittest.main()



from unittest import TestCase
import os
import pandas as pd
from wisdem.landbosse.model import SitePreparationCost
from wisdem.test.test_landbosse.test_WeatherDelay import generate_a_year
from wisdem.test.test_landbosse.test_filename_functions import landbosse_test_input_dir
import pytest

pd.set_option('display.width', 6000)
pd.set_option('display.max_columns', 20)

@pytest.mark.skip(reason="this does not pass")
class TestSitePreparationCost(TestCase):

    def setUp(self):
        self.input_dict = dict()
        self.project_name = 'Project_1'
        self.input_dict['road_distributed_wind'] = True #TODO add this to input file.

        #Inputs for calculate_road_properties():
        # self.input_dict['road_length'] = 108000   #TODO: Also add option to make road_length a user specified input.
        self.input_dict['road_width_ft'] = 10  # feet 10
        self.input_dict['road_thickness'] = 2  # inches
        self.input_dict['crane_width'] = 4.548  # metres 10.7 + shoulder width 1.5 m
        self.input_dict['num_access_roads'] = 5
        self.input_dict['num_turbines'] = 1
        self.input_dict['rsmeans'] = pd.read_csv(os.path.join(landbosse_test_input_dir(), 'rsmeans_proprietary.csv'))
        crew_data = os.path.join(landbosse_test_input_dir(), 'crews.csv')
        self.input_dict['crew'] = pd.read_csv(crew_data)
        crew_cost_data = os.path.join(landbosse_test_input_dir(), 'crew_price_proprietary.csv')
        self.input_dict['crew_cost'] = pd.read_csv(crew_cost_data)
        self.input_dict['duration_construction'] = 0.1 # months
        self.input_dict['construct_duration'] = 0.1 # months
        self.input_dict['fraction_new_roads'] = 1
        self.input_dict['road_quality'] = 0.6
        self.input_dict['rsmeans_per_diem'] = 144

        self.input_dict['road_length_adder_m'] = 5000   # 5km
        self.input_dict['fraction_new_roads'] = 0.33
        self.input_dict['road_quality'] = 0.3
        self.input_dict['rsmeans_per_diem'] = 0


        # Weather window:
        self.input_dict['num_delays'] = 7
        self.input_dict['avg_hours_per_delay'] = 20
        self.input_dict['std_dev_hours_per_delay'] = 5
        self.input_dict['delay_speed_m_per_s'] = 9
        self.input_dict['seed'] = 101
        self.input_dict['weather_window'] = generate_a_year(num_delays=self.input_dict['num_delays'],
                                                            avg_hours_per_delay=self.input_dict['avg_hours_per_delay'],
                                                            std_dev_hours_per_delay=self.input_dict[
                                                                'std_dev_hours_per_delay'],
                                                            delay_speed_m_per_s=self.input_dict['delay_speed_m_per_s'],
                                                            seed=self.input_dict['seed'])

        self.input_dict['start_delay_hours'] = 100
        self.input_dict['mission_time_hours'] = 50
        self.input_dict['critical_wind_speed_m_per_s'] = 8.0
        self.input_dict['wind_height_of_interest_m'] = 10
        self.input_dict['wind_shear_exponent'] = 0.25
        self.input_dict['season_construct'] = ['spring', 'summer']
        self.input_dict['time_construct'] = 'normal'
        self.input_dict['hour_day'] = {'long': 24, 'normal': 10}
        self.input_dict['operational_hrs_per_day'] = 10
        self.input_dict['critical_speed_non_erection_wind_delays_m_per_s'] = 15
        self.input_dict['critical_height_non_erection_wind_delays_m'] = 10

        #Inputs for calculate_costs():
        material_price_csv = os.path.join(landbosse_test_input_dir(), 'material_price_proprietary.csv')
        self.input_dict['material_price'] = pd.read_csv(material_price_csv)
        self.input_dict['turbine_rating_MW'] = 0.02  # 20 kW Turbine
        self.input_dict['overtime_multiplier'] = 1.4 # multiplier for labor overtime rates due to working 60 hr/wk rather than 40 hr/wk
        self.input_dict['rotor_diameter_m'] = 7
        self.input_dict['turbine_spacing_rotor_diameters'] = 5
        self.input_dict['critical_height_non_erection_wind_delays_m'] = 10
        self.input_dict['site_prep_area_m2'] = 0


        self.output_dict = dict()

    def test_SitePreparationCostModule(self):
        """
        Black box test to check whether module is ran successfully or not
        """
        run_RoadsCost = SitePreparationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        trial_run = run_RoadsCost.run_module()


        if trial_run[0] == 0 :
            print('\n\n================== MODULE EXECUTION SUCCESS =========================\n')
            print(' SitePreparationCost module ran successfully. See the list of inputs'
                  '\n and outputs below used by the module in its calculations:')
            print( '\n=====================================================================\n')

        elif trial_run[0] == 1 :
            print('\n\n====================================== MODULE EXECUTION FAILURE ======================================\n')
            print(' > SitePreparationCost module failed to run successfully. Error detected: ', trial_run[1],
                  '\n > Scroll below to see detailed information about error encountered.'
                  '\n > See the list of inputs below used by the module in its calculations:')
            print('\n========================================================================================================\n')

        print('\nGiven below is the set of inputs fed into RoadsCost module:\n')
        for key, value in self.input_dict.items():
            if isinstance(value, pd.DataFrame):
                print(key, ':\n', value)
            else:
                print(key, ':', value)


        if trial_run[0] == 0:  #Only print outputs if module ran successfully.
            print('\nGiven below is the set of outputs calculated by the RoadsCost module:\n')
            for key, value in self.output_dict.items():
                if isinstance(value, pd.DataFrame):
                    print('\nNow printing DataFrame ->', key, ':\n', value)
                else:
                    print(key, ':', value)

    def test_calculate_road_properties(self):
        road_properties = SitePreparationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        road_properties.run_module()

    def test_estimate_construction_time(self):
        construction_time = SitePreparationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        construction_time.run_module()

    def test_calculate_weather_delay(self):
        """
        Tests calculate_weather_delay()

        """

        weatherDelay = SitePreparationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        weatherDelay.run_module()

    def test_calculate_costs(self):
        calculate_costs = SitePreparationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        calculate_costs.run_module()



import os
import unittest

import numpy as np
import numpy.testing as npt

import wisdem.rotorse.rotor_structure as rs
from wisdem.commonse import gravity

ARCHIVE1 = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + "nrel5mw_test.npz"
ARCHIVE2 = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + "nrel5mw_test2.npz"


class TestRS(unittest.TestCase):
    def testBladeCurvature(self):
        inputs = {}
        outputs = {}
        npts = 101
        myzero = np.zeros(npts)
        myone = np.ones(npts)
        options = {}
        options["WISDEM"] = {}
        options["WISDEM"]["RotorSE"] = {}
        options["WISDEM"]["RotorSE"]["n_span"] = npts

        myobj = rs.BladeCurvature(modeling_options=options)

        # Straight blade: Z is 'r'
        inputs["r"] = np.linspace(0, 100, npts)
        inputs["precurve"] = myzero
        inputs["presweep"] = myzero
        inputs["precone"] = 0.0
        inputs["Rhub"] = 1.0
        inputs["blade_span_cg"] = 0.5 * inputs["r"].max()
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["3d_curv"], myzero)
        npt.assert_equal(outputs["x_az"], myzero)
        npt.assert_equal(outputs["y_az"], myzero)
        npt.assert_equal(outputs["z_az"], inputs["r"])
        npt.assert_equal(outputs["blades_cg_hubcc"], 0.0)

        # Some coning: Z is 'r'
        inputs["precone"] = 3.0
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["3d_curv"], 3 * myone)
        npt.assert_equal(outputs["x_az"], myzero)
        npt.assert_equal(outputs["y_az"], myzero)
        npt.assert_equal(outputs["z_az"], inputs["r"])
        npt.assert_equal(outputs["blades_cg_hubcc"], 51*np.sin(np.deg2rad(3)))

        # Some curve: X is 'flap'
        inputs["precurve"] = np.linspace(0, 1, npts)
        inputs["precone"] = 0.0
        myobj.compute(inputs, outputs)
        cone = -np.rad2deg(np.arctan(inputs["precurve"] / (inputs["r"] + 1e-20)))
        cone[0] = cone[1]
        npt.assert_almost_equal(outputs["3d_curv"], cone)
        npt.assert_equal(outputs["x_az"], inputs["precurve"])
        npt.assert_equal(outputs["y_az"], myzero)
        npt.assert_equal(outputs["z_az"], inputs["r"])
        npt.assert_almost_equal(outputs["blades_cg_hubcc"], 51*np.sin(np.deg2rad(cone[50])))

        # Some curve: Y is 'edge'
        inputs["precurve"] = myzero
        inputs["presweep"] = np.linspace(0, 1, npts)
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["3d_curv"], myzero)
        npt.assert_equal(outputs["x_az"], myzero)
        npt.assert_equal(outputs["y_az"], inputs["presweep"])
        npt.assert_equal(outputs["z_az"], inputs["r"])
        npt.assert_equal(outputs["blades_cg_hubcc"], np.zeros(3))

        # Some curve and sweep
        inputs["precurve"] = np.linspace(0, 2, npts)
        inputs["presweep"] = np.linspace(0, 1, npts)
        inputs["precone"] = 0.0
        myobj.compute(inputs, outputs)
        cone = -np.rad2deg(np.arctan(inputs["precurve"] / (inputs["r"] + 1e-20)))
        cone[0] = cone[1]
        npt.assert_almost_equal(outputs["3d_curv"], cone)
        npt.assert_equal(outputs["x_az"], inputs["precurve"])
        npt.assert_equal(outputs["y_az"], inputs["presweep"])
        npt.assert_equal(outputs["z_az"], inputs["r"])
        npt.assert_almost_equal(outputs["blades_cg_hubcc"], 51*np.sin(np.deg2rad(cone[50])))

    def testTotalBladeLoads(self):
        inputs = {}
        outputs = {}
        npts = 101
        myzero = np.zeros(npts)
        myone = np.ones(npts)
        options = {}
        options["WISDEM"] = {}
        options["WISDEM"]["RotorSE"] = {}
        options["WISDEM"]["RotorSE"]["n_span"] = npts

        myobj = rs.TotalBladeLoads(modeling_options=options)

        # Pass through
        inputs["r"] = np.linspace(0, 100, npts)
        inputs["aeroloads_Px"] = 10.0 * myone
        inputs["aeroloads_Py"] = 5.0 * myone
        inputs["aeroloads_Pz"] = 1.0 * myone
        inputs["aeroloads_Omega"] = 0.0
        inputs["aeroloads_pitch"] = 0.0
        inputs["aeroloads_azimuth"] = 0.0
        inputs["theta"] = myzero
        inputs["tilt"] = 0.0
        inputs["3d_curv"] = myzero
        inputs["z_az"] = inputs["r"]
        inputs["rhoA"] = myzero  # 20.0*myone
        inputs["dynamicFactor"] = 1.0
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["Px_af"], inputs["aeroloads_Px"])
        npt.assert_equal(outputs["Py_af"], inputs["aeroloads_Py"])
        npt.assert_equal(outputs["Pz_af"], inputs["aeroloads_Pz"])

        # With theta and pitch
        inputs["aeroloads_pitch"] = 20.0
        inputs["theta"] = 70 * myone
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px_af"], -inputs["aeroloads_Py"])
        npt.assert_almost_equal(outputs["Py_af"], inputs["aeroloads_Px"])
        npt.assert_equal(outputs["Pz_af"], inputs["aeroloads_Pz"])

        # With gravity
        inputs["aeroloads_pitch"] = 0.0
        inputs["theta"] = myzero
        inputs["rhoA"] = 20.0 * myone
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px_af"], inputs["aeroloads_Px"])
        npt.assert_almost_equal(outputs["Py_af"], inputs["aeroloads_Py"])
        npt.assert_equal(outputs["Pz_af"], inputs["aeroloads_Pz"] - 20 * gravity)

        # With gravity, az 180
        inputs["aeroloads_azimuth"] = 180.0
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px_af"], inputs["aeroloads_Px"])
        npt.assert_almost_equal(outputs["Py_af"], inputs["aeroloads_Py"])
        npt.assert_equal(outputs["Pz_af"], inputs["aeroloads_Pz"] + 20 * gravity)

        # With gravity, az 90
        inputs["aeroloads_azimuth"] = 90.0
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px_af"], inputs["aeroloads_Px"])
        npt.assert_almost_equal(outputs["Py_af"], inputs["aeroloads_Py"] - 20 * gravity)
        npt.assert_almost_equal(outputs["Pz_af"], inputs["aeroloads_Pz"])

        # With centrifical
        inputs["aeroloads_Omega"] = 5.0
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px_af"], inputs["aeroloads_Px"])
        npt.assert_almost_equal(outputs["Py_af"], inputs["aeroloads_Py"] - 20 * gravity)
        npt.assert_almost_equal(outputs["Pz_af"], inputs["aeroloads_Pz"] + 20 * inputs["r"] * (5 * 2 * np.pi / 60) ** 2)

    def testRunFrame3DD(self):
        inputs = {}
        outputs = {}

        nrel5mw = np.load(ARCHIVE1)
        for k in nrel5mw.files:
            inputs[k] = nrel5mw[k]

        npts = len(inputs["r"])
        nfreq = 10
        options = {}
        options["WISDEM"] = {}
        options["WISDEM"]["RotorSE"] = {}
        options["WISDEM"]["RotorSE"]["n_span"] = npts
        options["WISDEM"]["RotorSE"]["n_freq"] = nfreq

        myobj = rs.RunFrame3DD(modeling_options=options)
        myobj.n_span = npts
        myobj.n_freq = nfreq
        myobj.compute(inputs, outputs)

        # Truth values
        dx = [-0.00000000e+00,  1.64638232e-03,  6.79572518e-03,  1.57057110e-02,
        2.91434108e-02,  4.91143389e-02,  7.77022110e-02,  1.15994210e-01,
        1.64492394e-01,  2.23728011e-01,  2.94236207e-01,  3.76468424e-01,
        4.70954179e-01,  5.78649328e-01,  7.00809412e-01,  8.38646276e-01,
        9.93321364e-01,  1.16597533e+00,  1.35778027e+00,  1.56975422e+00,
        1.80318407e+00,  2.05880282e+00,  2.33562566e+00,  2.63177288e+00,
        2.94510570e+00,  3.27320110e+00,  3.61333244e+00,  3.96258446e+00,
        4.31781570e+00,  4.67663896e+00]
        dy = [ 0.00000000e+00, -3.74477007e-04, -1.87858152e-03, -5.72303746e-03,
       -1.40296446e-02, -2.91760415e-02, -5.25470603e-02, -8.42038955e-02,
       -1.23721261e-01, -1.70836582e-01, -2.25359315e-01, -2.87027856e-01,
       -3.55549679e-01, -4.30734595e-01, -5.12432617e-01, -6.00416367e-01,
       -6.94413956e-01, -7.94115178e-01, -8.99164661e-01, -1.00909604e+00,
       -1.12345396e+00, -1.24175723e+00, -1.36337226e+00, -1.48764046e+00,
       -1.61394215e+00, -1.74170551e+00, -1.87042453e+00, -1.99968813e+00,
       -2.12920515e+00, -2.25881481e+00]
        dz = [-0.00000000e+00, -6.16448640e-05, -1.46019534e-04, -2.32287182e-04,
       -3.21439626e-04, -4.16434884e-04, -5.11204867e-04, -6.03366938e-04,
       -6.92726810e-04, -7.79183257e-04, -8.62353953e-04, -9.41712542e-04,
       -1.01708823e-03, -1.08989853e-03, -1.15986208e-03, -1.22711716e-03,
       -1.29165787e-03, -1.35337047e-03, -1.41204679e-03, -1.46642332e-03,
       -1.51832829e-03, -1.56716073e-03, -1.61256368e-03, -1.65439674e-03,
       -1.69245182e-03, -1.72635996e-03, -1.75550444e-03, -1.77888069e-03,
       -1.79390194e-03, -1.80321598e-03]

        freqs = [0.83325001,  1.22280141,  2.9647584 ,  4.09217506,  6.86529582,
        8.83012966,  9.47967717, 12.57008964, 14.66188868, 16.55302485]

        npt.assert_almost_equal(outputs["dx"], dx, decimal=2)
        npt.assert_almost_equal(outputs["dy"], dy, decimal=2)
        npt.assert_almost_equal(outputs["dz"], dz, decimal=3)  # Very small numbers, so no precision
        npt.assert_almost_equal(outputs["freqs"], freqs, decimal=0)

    def testComputeStrains(self):
        inputs = {}
        outputs = {}

        nrel5mw = np.load(ARCHIVE1)
        for k in nrel5mw.files:
            inputs[k.replace("_strain_", "_")] = nrel5mw[k]

        nrel5mw = np.load(ARCHIVE2)
        for k in nrel5mw.files:
            inputs[k.replace("_strain_", "_")] = nrel5mw[k]

        npts = len(inputs["EA"])
        inputs["chord"] = np.ones(npts)

        options = {}
        options["WISDEM"] = {}
        options["WISDEM"]["RotorSE"] = {}
        options["WISDEM"]["RotorSE"]["n_span"] = npts

        myobj = rs.ComputeStrains(modeling_options=options)
        myobj.n_span = npts
        myobj.compute(inputs, outputs)

        # Truth values
        strainU_spar = [
            -9.60168580e-04,
            -1.36569964e-03,
            -1.24469019e-03,
            -1.28357551e-03,
            -1.66756907e-03,
            -1.93501003e-03,
            -1.96953848e-03,
            -1.93469820e-03,
            -1.91977182e-03,
            -1.87978271e-03,
            -1.82931324e-03,
            -1.79372052e-03,
            -1.79979393e-03,
            -1.81043159e-03,
            -1.80173183e-03,
            -1.77805206e-03,
            -1.74193967e-03,
            -1.69332271e-03,
            -1.63809273e-03,
            -1.57934205e-03,
            -1.51158159e-03,
            -1.38263396e-03,
            -1.23607178e-03,
            -1.07654944e-03,
            -9.03175151e-04,
            -7.15939331e-04,
            -5.20175274e-04,
            -3.25702417e-04,
            -1.74039663e-04,
            1.74416274e-08,
        ]
        strainL_spar = [
            9.15570014e-04,
            1.29000729e-03,
            1.16792215e-03,
            1.12951553e-03,
            1.39061424e-03,
            1.59834244e-03,
            1.63691918e-03,
            1.61549834e-03,
            1.60998340e-03,
            1.58649770e-03,
            1.55285394e-03,
            1.53887355e-03,
            1.57102695e-03,
            1.60654520e-03,
            1.61866703e-03,
            1.60814193e-03,
            1.57992027e-03,
            1.53608654e-03,
            1.48582402e-03,
            1.43671112e-03,
            1.37668380e-03,
            1.25754211e-03,
            1.12136865e-03,
            9.72322322e-04,
            8.10394743e-04,
            6.36106067e-04,
            4.55315076e-04,
            2.78529226e-04,
            1.43294401e-04,
            -1.50254249e-08,
        ]
        strainU_te = [-1.13026497e-04, -2.92410541e-04,  6.98603371e-05,  7.83552246e-04,
        1.03320930e-03,  1.11029244e-03,  1.07605717e-03,  1.00197550e-03,
        9.79127686e-04,  9.35549026e-04,  8.73496045e-04,  8.04623727e-04,
        7.46684889e-04,  7.13059487e-04,  6.92932872e-04,  6.66579608e-04,
        6.39927843e-04,  6.11292545e-04,  5.76948672e-04,  5.47664578e-04,
        5.09787645e-04,  4.35963414e-04,  3.45126223e-04,  2.70740910e-04,
        2.09983984e-04,  1.56888815e-04,  1.08873980e-04,  6.50476231e-05,
        3.44554780e-05, -3.57317878e-09]
        strainL_te = [-1.13026497e-04,  3.26449732e-04,  5.89457792e-04,  1.10079177e-03,
        1.24723249e-03,  1.26863500e-03,  1.21717638e-03,  1.13461525e-03,
        1.11141359e-03,  1.06753695e-03,  1.00345800e-03,  9.32406960e-04,
        8.77049612e-04,  8.49694704e-04,  8.40210719e-04,  8.34931478e-04,
        8.33043510e-04,  8.26570472e-04,  8.07511816e-04,  7.88325999e-04,
        7.48216769e-04,  6.60545168e-04,  5.48461897e-04,  4.40175017e-04,
        3.38218694e-04,  2.43921743e-04,  1.59467146e-04,  8.86770957e-05,
        4.62181259e-05, -3.57317878e-09]

        npt.assert_almost_equal(outputs["strainU_spar"], strainU_spar, decimal=3)
        npt.assert_almost_equal(outputs["strainL_spar"], strainL_spar, decimal=3)
        npt.assert_almost_equal(outputs["strainU_te"], strainU_te, decimal=3)
        npt.assert_almost_equal(outputs["strainL_te"], strainL_te, decimal=3)

        x_spar = inputs["xu_spar"][0]
        y_spar = inputs["yu_spar"][0]
        A = inputs["A"][0]
        E = inputs["EA"][0] / A
        npt.assert_almost_equal(outputs["axial_root_sparU_load2stress"][:2], 0.0)
        npt.assert_almost_equal(outputs["axial_root_sparU_load2stress"][5], 0.0)
        npt.assert_almost_equal(outputs["axial_root_sparU_load2stress"][2], -1.0 / A, decimal=3)
        npt.assert_almost_equal(outputs["axial_root_sparU_load2stress"][3], -E * y_spar / inputs["EI11"][0], decimal=2)
        npt.assert_almost_equal(outputs["axial_root_sparU_load2stress"][4], E * x_spar / inputs["EI22"][0], decimal=3)
        x_spar = inputs["xl_spar"][0]
        y_spar = inputs["yl_spar"][0]
        npt.assert_almost_equal(outputs["axial_root_sparL_load2stress"][:2], 0.0)
        npt.assert_almost_equal(outputs["axial_root_sparL_load2stress"][5], 0.0)
        npt.assert_almost_equal(outputs["axial_root_sparL_load2stress"][2], -1.0 / A, decimal=3)
        npt.assert_almost_equal(outputs["axial_root_sparL_load2stress"][3], -E * y_spar / inputs["EI11"][0], decimal=2)
        npt.assert_almost_equal(outputs["axial_root_sparL_load2stress"][4], E * x_spar / inputs["EI22"][0], decimal=3)

        k = np.argmax(inputs["chord"])
        x_te = inputs["xu_te"][k]
        y_te = inputs["yu_te"][k]
        A = inputs["A"][k]
        E = inputs["EA"][k] / A
        npt.assert_almost_equal(outputs["axial_maxc_teU_load2stress"][:2], 0.0)
        npt.assert_almost_equal(outputs["axial_maxc_teU_load2stress"][5], 0.0)
        npt.assert_almost_equal(outputs["axial_maxc_teU_load2stress"][2], -1.0 / A, decimal=3)
        npt.assert_almost_equal(outputs["axial_maxc_teU_load2stress"][3], -E * y_te / inputs["EI11"][k], decimal=2)
        npt.assert_almost_equal(outputs["axial_maxc_teU_load2stress"][4], E * x_te / inputs["EI22"][k], decimal=2)
        x_te = inputs["xl_te"][k]
        y_te = inputs["yl_te"][k]
        npt.assert_almost_equal(outputs["axial_maxc_teL_load2stress"][:2], 0.0)
        npt.assert_almost_equal(outputs["axial_maxc_teL_load2stress"][5], 0.0)
        npt.assert_almost_equal(outputs["axial_maxc_teL_load2stress"][2], -1.0 / A, decimal=3)
        npt.assert_almost_equal(outputs["axial_maxc_teL_load2stress"][3], -E * y_te / inputs["EI11"][k], decimal=2)
        npt.assert_almost_equal(outputs["axial_maxc_teL_load2stress"][4], E * x_te / inputs["EI22"][k], decimal=2)

    def testConstraints(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}

        npts = 101
        myzero = np.zeros(npts)
        myone = np.ones(npts)
        options = {}
        options["WISDEM"] = {}
        options["WISDEM"]["RotorSE"] = {}
        options["WISDEM"]["RotorSE"]["n_span"] = npts
        options["WISDEM"]["RotorSE"]["n_freq"] = 6
        options["WISDEM"]["RotorSE"]["gamma_freq"] = 1.1

        myopt = {}
        myopt["design_variables"] = {}
        myopt["design_variables"]["blade"] = {}
        myopt["design_variables"]["blade"]["structure"] = {}
        myopt["design_variables"]["blade"]["spar_cap_ss"] = {}
        myopt["design_variables"]["blade"]["spar_cap_ps"] = {}
        myopt["design_variables"]["blade"]["spar_cap_ss"]["n_opt"] = 3
        myopt["design_variables"]["blade"]["spar_cap_ps"]["n_opt"] = 3
        myopt["design_variables"]["blade"]["te_ss"] = {}
        myopt["design_variables"]["blade"]["te_ps"] = {}
        myopt["design_variables"]["blade"]["te_ss"]["n_opt"] = 3
        myopt["design_variables"]["blade"]["te_ps"]["n_opt"] = 3

        myobj = rs.DesignConstraints(modeling_options=options, opt_options=myopt)

        # Straight blade: Z is 'r'
        inputs["strainU_spar"] = np.linspace(0.4, 0.6, npts)
        inputs["strainL_spar"] = 0.6 * myone
        inputs["strainU_te"] = np.linspace(0.4, 0.6, npts)
        inputs["strainL_te"] = 0.6 * myone
        # inputs["min_strainU_spar"] = inputs["min_strainL_spar"] = 0.0
        inputs["max_strainU_spar"] = inputs["max_strainL_spar"] = 0.5
        inputs["max_strainU_te"] = inputs["max_strainL_te"] = 0.5
        inputs["s"] = np.linspace(0, 1, npts)
        inputs["s_opt_spar_cap_ss"] = inputs["s_opt_spar_cap_ps"] = np.array([0.0, 0.5, 1.0])
        inputs["s_opt_te_ss"] = inputs["s_opt_te_ps"] = np.array([0.0, 0.5, 1.0])
        inputs["rated_Omega"] = 10.0
        inputs["flap_mode_freqs"] = 0.6 * np.ones(3)
        inputs["edge_mode_freqs"] = 0.4 * np.ones(3)
        inputs["tors_mode_freqs"] = 5. * np.ones(3)
        discrete_inputs["blade_number"] = 3

        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["constr_max_strainU_spar"], np.array([0.8, 1.0, 1.2]))
        npt.assert_equal(outputs["constr_max_strainL_spar"], 1.2 * np.ones(3))
        npt.assert_equal(outputs["constr_max_strainU_te"], np.array([0.8, 1.0, 1.2]))
        npt.assert_equal(outputs["constr_max_strainL_te"], 1.2 * np.ones(3))
        npt.assert_almost_equal(outputs["constr_flap_f_margin"], 0.5 - 0.9 * 0.6)
        npt.assert_almost_equal(outputs["constr_edge_f_margin"], 1.1 * 0.4 - 0.5)


def suite():
    suite = [
        unittest.TestLoader().loadTestsFromTestCase(TestRS),
    ]
    return unittest.TestSuite(suite)


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



"""generator.py
Created by Latha Sethuraman, Katherine Dykes.
Copyright (c) NREL. All rights reserved.

Electromagnetic design based on conventional magnetic circuit laws
Structural design based on McDonald's thesis """

import numpy as np
import openmdao.api as om

import wisdem.drivetrainse.generator_models as gm

# ----------------------------------------------------------------------------------------------


class Constraints(om.ExplicitComponent):
    """
    Provides a material cost estimate for a PMSG _arms generator. Manufacturing costs are excluded.

    Parameters
    ----------
    u_allow_s : float, [m]

    u_as : float, [m]

    z_allow_s : float, [m]

    z_as : float, [m]

    y_allow_s : float, [m]

    y_as : float, [m]

    b_allow_s : float, [m]

    b_st : float, [m]

    u_allow_r : float, [m]

    u_ar : float, [m]

    y_allow_r : float, [m]

    y_ar : float, [m]

    z_allow_r : float, [m]

    z_ar : float, [m]

    b_allow_r : float, [m]

    b_arm : float, [m]

    TC1 : float, [m**3]

    TC2r : float, [m**3]

    TC2s : float, [m**3]

    B_g : float, [T]

    B_smax : float, [T]

    K_rad : float

    K_rad_LL : float

    K_rad_UL : float

    D_ratio : float

    D_ratio_LL : float

    D_ratio_UL : float


    Returns
    -------
    con_uas : float, [m]

    con_zas : float, [m]

    con_yas : float, [m]

    con_bst : float, [m]

    con_uar : float, [m]

    con_yar : float, [m]

    con_zar : float, [m]

    con_br : float, [m]

    TCr : float, [m**3]

    TCs : float, [m**3]

    con_TC2r : float, [m**3]

    con_TC2s : float, [m**3]

    con_Bsmax : float, [T]

    K_rad_L : float

    K_rad_U : float

    D_ratio_L : float

    D_ratio_U : float


    """

    def setup(self):
        self.add_input("u_allow_s", val=0.0, units="m")
        self.add_input("u_as", val=0.0, units="m")
        self.add_input("z_allow_s", val=0.0, units="m")
        self.add_input("z_as", val=0.0, units="m")
        self.add_input("y_allow_s", val=0.0, units="m")
        self.add_input("y_as", val=0.0, units="m")
        self.add_input("b_allow_s", val=0.0, units="m")
        self.add_input("b_st", val=0.0, units="m")
        self.add_input("u_allow_r", val=0.0, units="m")
        self.add_input("u_ar", val=0.0, units="m")
        self.add_input("y_allow_r", val=0.0, units="m")
        self.add_input("y_ar", val=0.0, units="m")
        self.add_input("z_allow_r", val=0.0, units="m")
        self.add_input("z_ar", val=0.0, units="m")
        self.add_input("b_allow_r", val=0.0, units="m")
        self.add_input("b_arm", val=0.0, units="m")
        self.add_input("TC1", val=0.0, units="m**3")
        self.add_input("TC2r", val=0.0, units="m**3")
        self.add_input("TC2s", val=0.0, units="m**3")
        self.add_input("B_g", val=0.0, units="T")
        self.add_input("B_smax", val=0.0, units="T")
        self.add_input("K_rad", val=0.0)
        self.add_input("K_rad_LL", val=0.0)
        self.add_input("K_rad_UL", val=0.0)
        self.add_input("D_ratio", val=0.0)
        self.add_input("D_ratio_LL", val=0.0)
        self.add_input("D_ratio_UL", val=0.0)

        self.add_output("con_uas", val=0.0, units="m")
        self.add_output("con_zas", val=0.0, units="m")
        self.add_output("con_yas", val=0.0, units="m")
        self.add_output("con_bst", val=0.0, units="m")
        self.add_output("con_uar", val=0.0, units="m")
        self.add_output("con_yar", val=0.0, units="m")
        self.add_output("con_zar", val=0.0, units="m")
        self.add_output("con_br", val=0.0, units="m")
        self.add_output("TCr", val=0.0, units="m**3")
        self.add_output("TCs", val=0.0, units="m**3")
        self.add_output("con_TC2r", val=0.0, units="m**3")
        self.add_output("con_TC2s", val=0.0, units="m**3")
        self.add_output("con_Bsmax", val=0.0, units="T")
        self.add_output("K_rad_L", val=0.0)
        self.add_output("K_rad_U", val=0.0)
        self.add_output("D_ratio_L", val=0.0)
        self.add_output("D_ratio_U", val=0.0)

    def compute(self, inputs, outputs):
        outputs["con_uas"] = inputs["u_allow_s"] - inputs["u_as"]
        outputs["con_zas"] = inputs["z_allow_s"] - inputs["z_as"]
        outputs["con_yas"] = inputs["y_allow_s"] - inputs["y_as"]
        outputs["con_bst"] = inputs["b_allow_s"] - inputs["b_st"]  # b_st={'units':'m'}
        outputs["con_uar"] = inputs["u_allow_r"] - inputs["u_ar"]
        outputs["con_yar"] = inputs["y_allow_r"] - inputs["y_ar"]
        outputs["con_TC2r"] = inputs["TC2s"] - inputs["TC1"]
        outputs["con_TC2s"] = inputs["TC2s"] - inputs["TC1"]
        outputs["con_Bsmax"] = inputs["B_g"] - inputs["B_smax"]
        outputs["con_zar"] = inputs["z_allow_r"] - inputs["z_ar"]
        outputs["con_br"] = inputs["b_allow_r"] - inputs["b_arm"]  # b_r={'units':'m'}
        outputs["TCr"] = inputs["TC2r"] - inputs["TC1"]
        outputs["TCs"] = inputs["TC2s"] - inputs["TC1"]
        outputs["K_rad_L"] = inputs["K_rad"] - inputs["K_rad_LL"]
        outputs["K_rad_U"] = inputs["K_rad"] - inputs["K_rad_UL"]
        outputs["D_ratio_L"] = inputs["D_ratio"] - inputs["D_ratio_LL"]
        outputs["D_ratio_U"] = inputs["D_ratio"] - inputs["D_ratio_UL"]


# ----------------------------------------------------------------------------------------------
class MofI(om.ExplicitComponent):
    """
    Compute moments of inertia.

    Parameters
    ----------
    R_out : float, [m]
        Outer radius
    stator_mass : float, [kg]
        Total rotor mass
    rotor_mass : float, [kg]
        Total rotor mass
    generator_mass : float, [kg]
        Actual mass
    len_s : float, [m]
        Stator core length

    Returns
    -------
    generator_I : numpy array[3], [kg*m**2]
        Moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass
    rotor_I : numpy array[3], [kg*m**2]
        Moments of Inertia for the rotor about its center of mass
    stator_I : numpy array[3], [kg*m**2]
        Moments of Inertia for the stator about its center of mass

    """

    def setup(self):
        self.add_input("R_out", val=0.0, units="m")
        self.add_input("stator_mass", val=0.0, units="kg")
        self.add_input("rotor_mass", val=0.0, units="kg")
        self.add_input("generator_mass", val=0.0, units="kg")
        self.add_input("len_s", val=0.0, units="m")

        self.add_output("generator_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("rotor_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("stator_I", val=np.zeros(3), units="kg*m**2")

    def compute(self, inputs, outputs):
        R_out = float(inputs["R_out"][0])
        Mass = float(inputs["generator_mass"][0])
        m_stator = float(inputs["stator_mass"][0])
        m_rotor = float(inputs["rotor_mass"][0])
        len_s = float(inputs["len_s"][0])

        I = np.zeros(3)
        I[0] = 0.50 * Mass * R_out**2
        I[1] = I[2] = 0.5 * I[0] + Mass * len_s**2 / 12.0
        outputs["generator_I"] = I
        coeff = m_stator / Mass if m_stator > 0.0 else 0.5
        outputs["stator_I"] = coeff * I
        coeff = m_rotor / Mass if m_rotor > 0.0 else 0.5
        outputs["rotor_I"] = coeff * I


# ----------------------------------------------------------------------------------------------


class Cost(om.ExplicitComponent):
    """
    Provides a material cost estimate for a PMSG _arms generator. Manufacturing costs are excluded.

    Parameters
    ----------
    C_Cu : float, [USD/kg]
        Specific cost of copper
    C_Fe : float, [USD/kg]
        Specific cost of magnetic steel/iron
    C_Fes : float, [USD/kg]
        Specific cost of structural steel
    C_PM : float, [USD/kg]
        Specific cost of Magnet
    Copper : float, [kg]
        Copper mass
    Iron : float, [kg]
        Iron mass
    mass_PM : float, [kg]
        Magnet mass
    Structural_mass : float, [kg]
        Structural mass

    Returns
    -------
    generator_cost : float, [USD]
        Total cost

    """

    def setup(self):
        # Specific cost of material by type
        self.add_input("C_Cu", val=0.0, units="USD/kg")
        self.add_input("C_Fe", val=0.0, units="USD/kg")
        self.add_input("C_Fes", val=0.0, units="USD/kg")
        self.add_input("C_PM", val=0.0, units="USD/kg")

        # Mass of each material type
        self.add_input("Copper", val=0.0, units="kg")
        self.add_input("Iron", val=0.0, units="kg")
        self.add_input("mass_PM", val=0.0, units="kg")
        self.add_input("Structural_mass", val=0.0, units="kg")

        # Outputs
        self.add_output("generator_cost", val=0.0, units="USD")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):
        Copper = inputs["Copper"]
        Iron = inputs["Iron"]
        mass_PM = inputs["mass_PM"]
        Structural_mass = inputs["Structural_mass"]
        C_Cu = inputs["C_Cu"]
        C_Fes = inputs["C_Fes"]
        C_Fe = inputs["C_Fe"]
        C_PM = inputs["C_PM"]

        # Industrial electricity rate $/kWh https://www.eia.gov/electricity/monthly/epm_table_grapher.php?t=epmt_5_6_a
        k_e = 0.064

        # Material cost ($/kg) and electricity usage cost (kWh/kg)*($/kWh) for the materials with waste fraction
        K_copper = Copper * (1.26 * C_Cu + 96.2 * k_e)
        K_iron = Iron * (1.21 * C_Fe + 26.9 * k_e)
        K_pm = mass_PM * (1.0 * C_PM + 79.0 * k_e)
        K_steel = Structural_mass * (1.21 * C_Fes + 15.9 * k_e)
        # Account for capital cost and labor share from BLS MFP by NAICS
        outputs["generator_cost"] = (K_copper + K_pm) / 0.619 + (K_iron + K_steel) / 0.684


# ----------------------------------------------------------------------------------------------


class PowerElectronicsEff(om.ExplicitComponent):
    """
    Compute representative efficiency of power electronics

    Parameters
    ----------
    machine_rating : float, [W]
        Machine rating
    shaft_rpm : numpy array[n_pc], [rpm]
        rated speed of input shaft (lss for direct, hss for geared)
    eandm_efficiency : numpy array[n_pc]
        Generator electromagnetic efficiency values (<1)

    Returns
    -------
    converter_efficiency : numpy array[n_pc]
        Converter efficiency values (<1)
    transformer_efficiency : numpy array[n_pc]
        Transformer efficiency values (<1)
    generator_efficiency : numpy array[n_pc]
        Full generato and power electronics efficiency values (<1)

    """

    def initialize(self):
        self.options.declare("n_pc", default=20)

    def setup(self):
        n_pc = self.options["n_pc"]

        self.add_input("machine_rating", val=0.0, units="W")
        self.add_input("shaft_rpm", val=np.zeros(n_pc), units="rpm")
        self.add_input("eandm_efficiency", val=np.zeros(n_pc))

        self.add_output("converter_efficiency", val=np.zeros(n_pc))
        self.add_output("transformer_efficiency", val=np.zeros(n_pc))
        self.add_output("generator_efficiency", val=np.zeros(n_pc))

    def compute(self, inputs, outputs):
        # Unpack inputs
        rating = inputs["machine_rating"]
        rpmData = inputs["shaft_rpm"]
        rpmRatio = rpmData / rpmData[-1]

        # This converter efficiency is from the APEEM Group in 2020
        # See Sreekant Narumanchi, Kevin Bennion, Bidzina Kekelia, Ramchandra Kotecha
        # Converter constants
        v_dc, v_dc0, c0, c1, c2, c3 = 6600, 6200, -2.1e-10, 1.2e-5, 1.46e-3, -2e-4
        p_ac0, p_dc0 = 0.99 * rating, rating
        p_s0 = 1e-3 * p_dc0

        # calculated parameters
        a = p_dc0 * (1.0 + c1 * (v_dc - v_dc0))
        b = p_s0 * (1.0 + c2 * (v_dc - v_dc0))
        c = c0 * (1.0 + c3 * (v_dc - v_dc0))

        # Converter efficiency
        p_dc = rpmRatio * p_dc0
        p_ac = (p_ac0 / (a - b) - c * (a - b)) * (p_dc - b) + c * ((p_dc - b) ** 2)
        conv_eff = p_ac / p_dc

        # Transformer loss model is P_loss = P_0 + a^2 * P_k
        # a is output power/rated
        p0, pk, rT = 16.0, 111.0, 5.0 / 3.0
        a = rpmRatio * (1 / rT)
        # This gives loss in kW, so need to convert to efficiency
        trans_eff = 1.0 - (p0 + a * a * pk) / (1e-3 * rating)

        # Store outputs
        outputs["converter_efficiency"] = conv_eff
        outputs["transformer_efficiency"] = trans_eff
        outputs["generator_efficiency"] = conv_eff * trans_eff * inputs["eandm_efficiency"]


# ----------------------------------------------------------------------------------------------
class Generator(om.Group):
    def initialize(self):
        genTypes = ["scig", "dfig", "eesg", "pmsg_arms", "pmsg_disc", "pmsg_outer"]
        self.options.declare("design", values=genTypes + [m.upper() for m in genTypes])
        self.options.declare("n_pc", default=20)

    def setup(self):
        genType = self.options["design"]
        n_pc = self.options["n_pc"]

        # ivc = om.IndepVarComp()
        # sivc = om.IndepVarComp()

        self.set_input_defaults("B_r", val=1.2, units="T")
        self.set_input_defaults("P_Fe0e", val=1.0, units="W/kg")
        self.set_input_defaults("P_Fe0h", val=4.0, units="W/kg")
        self.set_input_defaults("S_N", val=-0.002)
        self.set_input_defaults("alpha_p", val=0.5 * np.pi * 0.7)
        self.set_input_defaults("b_r_tau_r", val=0.45)
        self.set_input_defaults("b_ro", val=0.004, units="m")
        self.set_input_defaults("b_s_tau_s", val=0.45)
        self.set_input_defaults("b_so", val=0.004, units="m")
        self.set_input_defaults("cofi", val=0.85)
        self.set_input_defaults("freq", val=60, units="Hz")
        self.set_input_defaults("h_i", val=0.001, units="m")
        self.set_input_defaults("h_sy0", val=0.0)
        self.set_input_defaults("h_w", val=0.005, units="m")
        self.set_input_defaults("k_fes", val=0.9)
        self.set_input_defaults("k_fillr", val=0.7)
        self.set_input_defaults("k_fills", val=0.65)
        self.set_input_defaults("k_s", val=0.2)
        #self.set_input_defaults("m", val=3)
        self.set_input_defaults("mu_0", val=np.pi * 4e-7, units="m*kg/s**2/A**2")
        self.set_input_defaults("mu_r", val=1.06, units="m*kg/s**2/A**2")
        self.set_input_defaults("p", val=3.0)
        self.set_input_defaults("phi", val=np.deg2rad(90), units="rad")
        #self.set_input_defaults("q1", val=6)
        #self.set_input_defaults("q2", val=4)
        self.set_input_defaults("ratio_mw2pp", val=0.7)
        self.set_input_defaults("resist_Cu", val=1.8e-8 * 1.4, units="ohm/m")
        self.set_input_defaults("sigma", val=40e3, units="Pa")
        self.set_input_defaults("y_tau_p", val=1.0)
        self.set_input_defaults("y_tau_pr", val=10.0 / 12)

        # self.set_input_defaults('I_0', val=0.0, units='A')
        # self.set_input_defaults('d_r', val=0.0, units='m')
        # self.set_input_defaults('h_m', val=0.0, units='m')
        # self.set_input_defaults('h_0', val=0.0, units ='m')
        # self.set_input_defaults('h_s', val=0.0, units='m')
        # self.set_input_defaults('len_s', val=0.0, units='m')
        # self.set_input_defaults('n_r', val=0.0)
        # self.set_input_defaults('rad_ag', val=0.0, units='m')
        # self.set_input_defaults('t_wr', val=0.0, units='m')

        # self.set_input_defaults('n_s', val=0.0)
        # self.set_input_defaults('b_st', val=0.0, units='m')
        # self.set_input_defaults('d_s', val=0.0, units='m')
        # self.set_input_defaults('t_ws', val=0.0, units='m')

        # self.set_input_defaults('rho_Copper', val=0.0, units='kg*m**-3')
        # self.set_input_defaults('rho_Fe', val=0.0, units='kg*m**-3')
        # self.set_input_defaults('rho_Fes', val=0.0, units='kg*m**-3')
        # self.set_input_defaults('rho_PM', val=0.0, units='kg*m**-3')

        # self.set_input_defaults('C_Cu',  val=0.0, units='USD/kg')
        # self.set_input_defaults('C_Fe',  val=0.0, units='USD/kg')
        # self.set_input_defaults('C_Fes', val=0.0, units='USD/kg')
        # self.set_input_defaults('C_PM',  val=0.0, units='USD/kg')

        # if genType.lower() in ['pmsg_outer']:
        #    self.set_input_defaults('r_g',0.0, units ='m')
        #    self.set_input_defaults('N_c',0.0)
        #    self.set_input_defaults('b',0.0)
        #    self.set_input_defaults('c',0.0)
        #    self.set_input_defaults('E_p',0.0, units ='V')
        #    self.set_input_defaults('h_yr', val=0.0, units ='m')
        #    self.set_input_defaults('h_ys', val=0.0, units ='m')
        #    self.set_input_defaults('h_sr',0.0,units='m',desc='Structural Mass')
        #    self.set_input_defaults('h_ss',0.0, units ='m')
        #    self.set_input_defaults('t_r',0.0, units ='m')
        #    self.set_input_defaults('t_s',0.0, units ='m')

        #    self.set_input_defaults('u_allow_pcent',0.0)
        #    self.set_input_defaults('y_allow_pcent',0.0)
        #    self.set_input_defaults('z_allow_deg',0.0,units='deg')
        #    self.set_input_defaults('B_tmax',0.0, units='T')

        #    self.set_input_defaults('P_mech', 0.0, units='W')
        #    self.set_input_defaults('y_sh', units ='m')
        #    self.set_input_defaults('theta_sh', 0.0, units='rad')
        #    self.set_input_defaults('D_nose',0.0, units ='m')
        #    self.set_input_defaults('y_bd', units ='m')
        #    self.set_input_defaults('theta_bd', 0.0, units='rad')

        # if genType.lower() in ['eesg','pmsg_arms','pmsg_disc']:
        #    self.set_input_defaults('tau_p', val=0.0, units='m')
        #    self.set_input_defaults('h_ys',  val=0.0, units='m')
        #    self.set_input_defaults('h_yr',  val=0.0, units='m')
        #    self.set_input_defaults('b_arm',   val=0.0, units='m')

        # elif genType.lower() in ['scig','dfig']:
        #    self.set_input_defaults('B_symax', val=0.0, units='T')
        #    self.set_input_defaults('S_Nmax', val=-0.2)

        # if topLevelFlag:
        #    self.add_subsystem('ivc', ivc, promotes=['*'])

        #    self.set_input_defaults('machine_rating', 0.0, units='W')
        #    self.set_input_defaults('shaft_rpm', np.linspace(1.0, 10.0, n_pc), units='rpm')
        #    self.set_input_defaults('rated_torque', 0.0, units='N*m')
        #    self.set_input_defaults('D_shaft', val=0.0, units='m')
        self.set_input_defaults("E", val=210e9, units="Pa")
        self.set_input_defaults("G", val=81e9, units="Pa")
        #    self.add_subsystem('sivc', sivc, promotes=['*'])

        # Easy Poisson ratio assuming isotropic
        self.add_subsystem(
            "poisson", om.ExecComp("v = 0.5*E/G - 1.0", E={"units": "Pa"}, G={"units": "Pa"}), promotes=["*"]
        )

        # Add generator design component and cost
        if genType.lower() == "scig":
            mygen = gm.SCIG

        elif genType.lower() == "dfig":
            mygen = gm.DFIG

        elif genType.lower() == "eesg":
            mygen = gm.EESG

        elif genType.lower() == "pmsg_arms":
            mygen = gm.PMSG_Arms

        elif genType.lower() == "pmsg_disc":
            mygen = gm.PMSG_Disc

        elif genType.lower() == "pmsg_outer":
            mygen = gm.PMSG_Outer

        self.add_subsystem("generator", mygen(n_pc=n_pc), promotes=["*"])
        self.add_subsystem("mofi", MofI(), promotes=["*"])
        self.add_subsystem("gen_cost", Cost(), promotes=["*"])
        self.add_subsystem("constr", Constraints(), promotes=["*"])
        self.add_subsystem("eff", PowerElectronicsEff(n_pc=n_pc), promotes=["*"])



import os
import unittest

from wisdem.inputs import load_yaml, write_yaml
from wisdem.glue_code.runWISDEM import run_wisdem

test_dir = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))),
    "examples",
    "05_tower_monopile")

fname_wt_input = os.path.join(test_dir, "nrel5mw_tower.yaml")
fname_modeling_options = os.path.join(test_dir, "modeling_options.yaml")
fname_analysis_options = os.path.join(test_dir, "analysis_options.yaml")


class TestRegression(unittest.TestCase):
    def test_GA(self):
        analysis_options = load_yaml(fname_analysis_options)

        solver = "GA"
        analysis_options["driver"]["optimization"]["solver"] = solver
        analysis_options["driver"]["optimization"]["max_gen"] = 1

        new_fname_analysis_options = f"{fname_analysis_options.split('/')[-1].split('.')[0]}_{solver}.yaml"
        write_yaml(analysis_options, new_fname_analysis_options)

        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, new_fname_analysis_options
        )

    def test_nelder(self):
        analysis_options = load_yaml(fname_analysis_options)

        solver = "Nelder-Mead"
        analysis_options["driver"]["optimization"]["solver"] = solver
        analysis_options["driver"]["optimization"]["adaptive"] = True
        analysis_options["driver"]["optimization"]["max_iter"] = 1

        new_fname_analysis_options = f"{fname_analysis_options.split('/')[-1].split('.')[0]}_{solver}.yaml"
        write_yaml(analysis_options, new_fname_analysis_options)

        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, new_fname_analysis_options
        )

    def test_cobyla(self):
        analysis_options = load_yaml(fname_analysis_options)

        solver = "COBYLA"
        analysis_options["driver"]["optimization"]["solver"] = solver
        analysis_options["driver"]["optimization"]["rhobeg"] = 0.5
        analysis_options["driver"]["optimization"]["max_iter"] = 1

        new_fname_analysis_options = f"{fname_analysis_options.split('/')[-1].split('.')[0]}_{solver}.yaml"
        write_yaml(analysis_options, new_fname_analysis_options)

        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, new_fname_analysis_options
        )

    def test_ld_slsqp(self):
        analysis_options = load_yaml(fname_analysis_options)

        solver = "LD_SLSQP"
        analysis_options["driver"]["optimization"]["solver"] = solver
        analysis_options["driver"]["optimization"]["max_iter"] = 1

        new_fname_analysis_options = f"{fname_analysis_options.split('/')[-1].split('.')[0]}_{solver}.yaml"
        write_yaml(analysis_options, new_fname_analysis_options)

        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, new_fname_analysis_options
        )


if __name__ == "__main__":
    unittest.main()



# 0 ---------- (marker for docs)
import openmdao.api as om

from wisdem.nrelcsm.nrel_csm_cost_2015 import Turbine_CostsSE_2015

# 0 ---------- (marker for docs)

# 1 ---------- (marker for docs)
# OpenMDAO Problem instance
prob = om.Problem(reports=False)
prob.model = Turbine_CostsSE_2015(verbosity=True)
prob.setup()
# 1 ---------- (marker for docs)

# 2 ---------- (marker for docs)
# Initialize variables for NREL CSM
prob["machine_rating"] = 5000.0
prob["blade_number"] = 3
prob["crane"] = True
prob["main_bearing_number"] = 2
# 2 ---------- (marker for docs)

# 3 ---------- (marker for docs)
# Component masses
prob["blade_mass"] = 15751.48043042
prob["hub_mass"] = 37548.40498997
prob["pitch_system_mass"] = 9334.08947551
prob["spinner_mass"] = 973.0
prob["lss_mass"] = 20568.96284886
prob["main_bearing_mass"] = 2245.41649102
prob["gearbox_mass"] = 21875.0
prob["hss_mass"] = 994.7
prob["generator_mass"] = 14900.0
prob["bedplate_mass"] = 41765.26095285
prob["yaw_mass"] = 12329.96247921
prob["hvac_mass"] = 400.0
prob["cover_mass"] = 6836.69
prob["tower_mass"] = 182336.48057717
prob["transformer_mass"] = 11485.0
prob["platforms_mass"] = 8220.65761911
# 3 ---------- (marker for docs)

# 4 ---------- (marker for docs)
# Evaluate the model
prob.run_model()
# 4 ---------- (marker for docs)

# 5 ---------- (marker for docs)
# Print all intermediate inputs and outputs to the screen
prob.model.list_inputs(units=True)
prob.model.list_outputs(units=True)
# 5 ---------- (marker for docs)



import os
import openmdao.api as om
from wisdem.glue_code.gc_RunTools import PlotRecorder
import wisdem.inputs as sch

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file

analysis_options_files = ["analysis_options_no_opt.yaml",
                          "analysis_options_aero.yaml",
                          "analysis_options_struct.yaml",
                          "analysis_options_aerostruct.yaml",
                          "analysis_options_user.yaml",
                          ]

for a in analysis_options_files:
    fname_analysis_options = os.path.join(mydir, a)
    analysis_options = sch.load_analysis_yaml(fname_analysis_options)
    wt_opt = om.Problem(model=PlotRecorder(opt_options=analysis_options))
    wt_opt.setup(derivatives=False)
    wt_opt.run_model()


import os
import unittest
import importlib
from time import time
from pathlib import Path

thisdir = os.path.dirname(os.path.realpath(__file__))
# Find examples directory- outside the module path

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(thisdir)))
examples_dir = os.path.join(root_dir, "examples")
all_examples = Path(examples_dir).rglob("*.py") if os.path.exists(examples_dir) else []

# 02_ref turbines are regression tested in test_gluecode, no need to duplicate runtime
all_scripts = [
    "01_nrel_csm/costs",
    "01_nrel_csm/mass",
    "01_nrel_csm/mass_and_cost",
    "01_nrel_csm/parametric",
    "02_reference_turbines/nrel5mw_driver",
    "02_reference_turbines/iea10mw_driver",
    "02_reference_turbines/iea15mw_driver",
    "02_reference_turbines/iea22mw_driver",
    "02_reference_turbines/iea3p4mw_driver",
    "03_blade/blade_driver",
    "03_blade/load_log",
    "04_openmdao/betz_limit",
    "04_openmdao/sellar",
    "05_tower_monopile/monopile_direct",
    "05_tower_monopile/monopile_driver",
    "05_tower_monopile/tower_direct",
    "05_tower_monopile/tower_driver",
    "06_drivetrain/drivetrain_direct",
    "06_drivetrain/drivetrain_geared",
    "07_generator/dfig",
    "07_generator/eesg",
    "07_generator/pmsg_arms",
    "07_generator/pmsg_disc",
    "07_generator/pmsg_outer",
    "07_generator/scig",
    "08_plant_finance/example",
    "09_floating/mooring_opt",
    "09_floating/semi_only_driver",
    "09_floating/spar_only_driver",
    "09_floating/spar_opt",
    "09_floating/tlp_example",
    "09_floating/nrel5mw-spar_oc3_driver",
    "09_floating/nrel5mw-semi_oc4_driver",
    "09_floating/iea15mw_driver",
    "09_floating/iea22mw_driver",
    "10_ccblade/example",
    "10_ccblade/gradients",
    "10_ccblade/precurve",
    # 11/nrel-spar_oc3_user_mass_driver and user_elastic_driver covered by test_glue_code/test_gc_user_mass
    "11_user_custom/user_custom_opt",
    "11_user_custom/variable_override",
    "12_pyframe3dd/exB",
    "13_design_of_experiments/doe_with_openmdao",
    "13_design_of_experiments/doe_custom",
    "15_step_size_study/driver",
    "16_inverse_design/inverse_spar_design",
    "16_inverse_design/inverse_rotor_design",
    "17_jacket/jacket_driver",
    "18_rotor_tower_monopile/design_run",
    "19_rotor_drivetrain_tower/wisdem_driver"
]


def execute_script(fscript):
    # Go to location due to relative path use for airfoil files
    print("\n\n")
    print("NOW RUNNING:", fscript)
    print()
    fullpath = os.path.join(examples_dir, fscript + ".py")
    basepath = os.path.join(examples_dir, fscript.split("/")[0])
    os.chdir(basepath)

    # Get script/module name
    froot = fscript.split("/")[-1]

    # Use dynamic import capabilities
    # https://www.blog.pythonlibrary.org/2016/05/27/python-201-an-intro-to-importlib/
    print(froot, os.path.realpath(fullpath))
    spec = importlib.util.spec_from_file_location(froot, os.path.realpath(fullpath))
    mod = importlib.util.module_from_spec(spec)
    s = time()
    spec.loader.exec_module(mod)
    print(time() - s, "seconds to run")


class TestExamples(unittest.TestCase):
    def test_all_scripts(self):
        for ks, s in enumerate(all_scripts):
            with self.subTest(f"Running: {s}", i=ks):
                try:
                    execute_script(s)
                    self.assertTrue(True)
                except Exception:
                    self.assertEqual(s, "Success")


def suite():
    suite = [
        unittest.TestLoader().loadTestsFromTestCase(TestExamples),
    ]
    return unittest.TestSuite(suite)


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import sys

import numpy as np
import openmdao.api as om

import wisdem.commonse.utilities as util
import logging
logger = logging.getLogger("wisdem/weis")

class FindMaxTorque(om.ExplicitComponent):
    """
    Compute the maximum torque from an emergency stop at rated speed in a specified amount of time.

    Parameters
    ----------
    blades_I : numpy array[6], [kg*m**2]
        Moment of inertia for all of the blades combined in hub-aligned c.s.
    rated_rpm : float, [rpm]
        Rated speed of the turbine in rotations per minute
    stop_time : float, [s]
        Time required for the turbine rotor to come to a complete stop

    Returns
    -------
    max_torque : float, [N*m]
        Maximum torque seen during emergency stop during rated conditions

    """

    def setup(self):
        self.add_input("blades_I", val=np.zeros(6), units="kg*m**2")
        self.add_input("rated_rpm", val=0.0, units="rpm")
        self.add_input("stop_time", val=5.0, units="s")

        self.add_output("max_torque", val=0.0, units="N*m")

    def compute(self, inputs, outputs):
        omega = float(inputs["rated_rpm"][0]) * np.pi / 30.0
        outputs["max_torque"] = float(inputs["blades_I"][0]) * omega / float(inputs["stop_time"][0])


class HubShell(om.ExplicitComponent):
    """
    Size the wind turbine rotor hub, modeled as a spherical shell

    Parameters
    ----------
    flange_t2shell_t : float
        Ratio of flange thickness to shell thickness
    flange_OD2hub_D : float
        Ratio of flange outer diameter to hub diameter
    flange_ID2flange_OD : float
        Ratio of flange inner diameter to flange outer diameter
    rho : float, [kg/m**3]
        Density metal
    max_torque : float, [N*m]
        Max torque that the hub needs to resist (Mx in a hub aliged reference system)
    Xy : float, [Pa]
        Yield strength metal for hub (200MPa is a good value for SN Cast Iron
        GJS-350 for thick sections)
    hub_stress_concentration : float
        Stress concentration factor. Stress concentration occurs at all fillets,
        notches, lifting lugs, hatches and are accounted for by assigning a
        stress concentration factor
    gamma : float
        Design safety factor
    metal_cost : float, [USD/kg]
        Unit cost metal
    hub_diameter : float, [m]
        Outer diameter of the hub
    hub_shell_mass_user : float, [kg]
        User override of the mass of the hub shell, including the flanges

    Returns
    -------
    hub_mass : float, [kg]
        Total mass of the hub shell, including the flanges
    hub_cost : float, [USD]
        Cost of the hub shell, including flanges
    hub_cm : float, [m]
        Distance between hub/shaft flange and hub center of mass
    hub_I : numpy array[6], [kg*m**2]
        Total mass moment of inertia of the hub about its cm

    """

    def initialize(self):
        self.options.declare("gamma", default=2.0)

    def setup(self):
        # Inputs
        self.add_input("flange_t2shell_t", val=0.0)
        self.add_input("flange_OD2hub_D", val=0.0)
        self.add_input("flange_ID2flange_OD", val=0.0)
        self.add_input("rho", val=0.0, units="kg/m**3")
        self.add_input("max_torque", val=0.0, units="N*m")
        self.add_input("Xy", val=0.0, units="Pa")
        self.add_input("hub_stress_concentration", val=0.0)
        self.add_input("metal_cost", val=0.0, units="USD/kg")
        self.add_input("hub_diameter", val=0.0, units="m")
        self.add_input("blade_root_diameter", val=0.0, units="m")
        self.add_input("hub_in2out_circ", val=1.2)
        self.add_discrete_input("n_blades", val=3)
        self.add_input("hub_shell_mass_user", val=0.0, units="kg")

        # Outputs
        self.add_output("hub_mass", val=0.0, units="kg")
        self.add_output("hub_cost", val=0.0, units="USD")
        self.add_output("hub_cm", val=0.0, units="m")
        self.add_output("hub_I", val=np.zeros(6), units="kg*m**2")
        self.add_output("constr_hub_diameter", val=0.0, units="m")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # User diameter
        dsgn_hub_diam = inputs["hub_diameter"]

        # Estimate diameter of sphere based on blade root diameter
        min_hub_diam = (
            inputs["hub_in2out_circ"] * inputs["blade_root_diameter"] / np.sin(np.pi / discrete_inputs["n_blades"])
        )

        # Compute diameter and radius of the hub including safety factor
        dsgn_hub_rad = 0.5 * dsgn_hub_diam

        # Determine max stress
        stress_allow_pa = inputs["Xy"] / (inputs["hub_stress_concentration"] * self.options["gamma"])

        # Size shell thickness, assuming max torsional stress from torque can not exceed design allowable stress, and solving, torsional stress (t=Tr/J), and polar moment of inertia (J=PI/32(Do^4-Di^4), for thickness.
        sph_hub_shell_thick = (
            (
                (dsgn_hub_diam**4.0 - 32.0 / np.pi * inputs["max_torque"] * dsgn_hub_rad / stress_allow_pa)
                ** (1.0 / 4.0)
            )
            - dsgn_hub_diam
        ) / (-2.0)

        # Compute volume and mass of the shell
        sph_hub_vol = 4.0 / 3.0 * np.pi * (dsgn_hub_rad**3.0 - (dsgn_hub_rad - sph_hub_shell_thick) ** 3.0)
        sph_hub_mass = sph_hub_vol * inputs["rho"]

        # Assume outer (OD) and inner diameter (ID) of the flanges based on hub diameter
        main_flange_OD = dsgn_hub_diam * inputs["flange_OD2hub_D"]
        main_flange_ID = main_flange_OD * inputs["flange_ID2flange_OD"]

        # Compute thickness, volume, and mass of main flange
        main_flange_thick = inputs["flange_t2shell_t"] * sph_hub_shell_thick
        main_flange_vol = np.pi * main_flange_thick * ((main_flange_OD / 2.0) ** 2.0 - (main_flange_ID / 2.0) ** 2.0)
        main_flange_mass = main_flange_vol * inputs["rho"]

        # Sum masses flange and hub
        hub_mass_user = float(inputs["hub_shell_mass_user"][0])
        hub_mass = main_flange_mass + sph_hub_mass if hub_mass_user == 0.0 else hub_mass_user

        # Compute cost
        hub_cost = hub_mass * inputs["metal_cost"]

        # Compute distance between hub/shaft flange and hub center of mass
        hub_cm = (main_flange_mass * main_flange_thick * 0.5 + sph_hub_mass * dsgn_hub_rad) / (
            main_flange_mass + sph_hub_mass
        )

        # Assign values to openmdao outputs
        outputs["hub_mass"] = hub_mass
        outputs["hub_cost"] = hub_cost
        outputs["hub_cm"] = hub_cm
        outputs["hub_I"] = np.r_[
            (2.0 / 3.0) * hub_mass * (0.5 * dsgn_hub_diam) ** 2 * np.ones(3), np.zeros(3)
        ]  # Spherical shell
        outputs["constr_hub_diameter"] = dsgn_hub_diam - min_hub_diam  # Should be > 0


class Spinner(om.ExplicitComponent):
    """
    Size the wind turbine rotor hub spinner, modeled as a spherical shell that wraps the hub

    Parameters
    ----------
    n_blades : integer
        Number of rotor blades
    n_front_brackets : integer
        Number of front spinner brackets
    n_rear_brackets : integer
        Number of rear spinner brackets
    blade_root_diameter : float, [m]
        Outer diameter of blade root
    hub_diameter : float, [m]
        Outer diameter of the hub
    clearance_hub_spinner : float, [m]
        Clearance between spinner and hub
    spin_hole_incr : float
        Ratio between access hole diameter in the spinner and blade root diameter.
        Typical value 1.2
    spinner_gust_ws : float, [m/s]
        Extreme gust wind speed
    composite_Xt : float, [Pa]
        Tensile strength of the composite material of the shell.
        A glass CFM (continuous fiber mat) is often used.
    composite_rho : float, [kg/m**3]
        Density of composite of the shell
    Xy : float, [Pa]
        Yield strength metal
    metal_rho : float, [kg/m**3]
        Density metal
    composite_cost : float, [USD/kg]
        Unit cost composite of the shell
    metal_cost : float, [USD/kg]
        Unit cost metal
    spinner_mass_user : float, [m]
        User override of spinner mass

    Returns
    -------
    spinner_mass : float, [m]
        Total mass of the spinner
    spinner_diameter : float, [kg]
        Outer diameter of the spinner
    spinner_cost : float, [kg]
        Cost of the spinner
    spinner_cm : float, [m]
        Distance between center of mass of the spinner and main flange
    spinner_I : numpy array[6], [kg*m**2]
        Total mass moment of inertia of the spinner about its cm

    """

    def initialize(self):
        self.options.declare("gamma", default=1.5)

    def setup(self):
        # Inputs
        self.add_discrete_input("n_blades", val=0)
        self.add_discrete_input("n_front_brackets", val=0)
        self.add_discrete_input("n_rear_brackets", val=0)
        self.add_input("blade_root_diameter", val=0.0, units="m")
        self.add_input("hub_diameter", val=0.0, units="m")
        self.add_input("clearance_hub_spinner", val=0.0, units="m")
        self.add_input("spin_hole_incr", val=0.0)
        self.add_input("spinner_gust_ws", val=0.0, units="m/s")
        self.add_input("composite_Xt", val=0.0, units="Pa")
        # self.add_input('composite_SF',              val = 0.0)
        self.add_input("composite_rho", val=0.0, units="kg/m**3")
        self.add_input("Xy", val=0.0, units="Pa")
        # self.add_input('metal_SF',                  val = 0.0)
        self.add_input("metal_rho", val=0.0, units="kg/m**3")
        self.add_input("composite_cost", val=0.0, units="USD/kg")
        self.add_input("metal_cost", val=0.0, units="USD/kg")
        self.add_input("spinner_mass_user", val=0.0, units="kg")

        # Outputs
        self.add_output("spinner_diameter", val=0.0, units="m")
        self.add_output("spinner_mass", val=0.0, units="kg")
        self.add_output("spinner_cost", val=0.0, units="kg")
        self.add_output("spinner_cm", val=0.0, units="m")
        self.add_output("spinner_I", val=np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Increase hub diameter by twice the clearance between hub and spinner. This gives us the spherical spinner diameter, radius, and circumference
        sph_spin_diam = inputs["hub_diameter"] + (2.0 * inputs["clearance_hub_spinner"])
        sph_spin_rad = 0.5 * sph_spin_diam
        sph_spin_circ = np.pi * sph_spin_diam
        dsgn_hub_circ = np.pi * inputs["hub_diameter"]

        # Compute the width of the panel between blade cutouts
        spin_panel_width = (sph_spin_circ - dsgn_hub_circ) / 3.0

        # Compute the access hole diameter given blade root diameter
        spin_acc_hole_diam = inputs["blade_root_diameter"] * inputs["spin_hole_incr"]

        # Estimate thrust pressure on spinner given a wind speed and a multiplicative factor
        extr_gust_dsgn_pressure = 0.5 * 1.225 * (inputs["spinner_gust_ws"] ** 2.0) * self.options["gamma"]

        # Compute max allowable tensile strength of composite and max allowable yield strength metal given material properties and safety factors
        allow_tensile_strength = inputs["composite_Xt"]  # / inputs['composite_SF']
        allow_yield_strength = inputs["Xy"]  # / inputs['metal_SF']

        # Estimate thickness of the shell of the spinner. Stress equation for a flat plate with simply supported edges, with a load equal to the extreme gust pressure load.
        # The equation is [Stress=(.75*P*b^2)/(t^2*(1.61*(b/a)^3 +1)).  See Roarks for reference.  Shell is curved and not flat but simplifying for calculation purposes.
        spin_shell_thickness = np.sqrt(
            (0.75 * extr_gust_dsgn_pressure * spin_panel_width**2.0)
            / (allow_tensile_strength * (1.61 * (spin_panel_width / sph_spin_diam) ** 3.0 + 1.0))
        )

        # Compute volume and mass of the spinner shell
        spin_shell_volume = (
            (4.0 / 3.0) * np.pi * (sph_spin_rad**3.0 - ((sph_spin_diam - 2.0 * spin_shell_thickness) / 2.0) ** 3.0)
        )
        spin_shell_mass = spin_shell_volume * inputs["composite_rho"]

        # Estimate area, volume, and mass of the spherical caps that are removed because of blade access
        # if condition to avoid error
        if spin_acc_hole_diam / 2.0 < sph_spin_rad:
            sph_cap_area = (
                2.0
                * np.pi
                * sph_spin_rad
                * (sph_spin_rad - np.sqrt(sph_spin_rad**2.0 - (spin_acc_hole_diam / 2.0) ** 2.0))
            )
        else:
            sph_cap_area = 0. 
            logger.warning('Blade root chord is excessive! Please reduce it, or alternatively increase the hub diameter.')
        sph_caps_volume = discrete_inputs["n_blades"] * sph_cap_area * spin_shell_thickness
        sph_caps_mass = sph_caps_volume * inputs["composite_rho"]

        # Estimate main flange diameter, area, volume, and mass
        main_flange_diam = 0.6 * inputs["hub_diameter"]
        main_flange_area = (
            2.0 * np.pi * sph_spin_rad * (sph_spin_rad - np.sqrt(sph_spin_rad**2.0 - (main_flange_diam / 2.0) ** 2.0))
        )
        main_flange_volume = main_flange_area * spin_shell_thickness
        main_flange_mass = main_flange_volume * inputs["composite_rho"]
        spin_shell_mass = spin_shell_mass - sph_caps_mass - main_flange_mass

        # Compute frontal area of spherical spinner
        spin_frontal_area = np.pi * (sph_spin_diam**2.0) / 4.0

        # Compute load given frontal area
        frontal_gust_load = spin_frontal_area * extr_gust_dsgn_pressure

        # Compute load on single bracket
        bracket_load = frontal_gust_load / (discrete_inputs["n_front_brackets"] + discrete_inputs["n_rear_brackets"])

        # Compute bending moment on bracket
        bracket_bending_moment = bracket_load * inputs["clearance_hub_spinner"]

        # Assume bracket width is half of the spinner panel width
        bracket_width = spin_panel_width / 2.0

        # Compute bracket thickness given loads and metal properties
        bracket_thickness = np.sqrt((6.0 * bracket_bending_moment) / (bracket_width * allow_yield_strength))

        # Assume flang is 25% of bracket length
        bracket_flange_length = inputs["clearance_hub_spinner"] * 0.25

        # Compute bracket volume and mass and total mass of all brackets
        bracket_volume = (
            (inputs["clearance_hub_spinner"] + bracket_flange_length + bracket_flange_length)
            * bracket_width
            * bracket_thickness
        )
        bracket_mass = bracket_volume * inputs["metal_rho"]
        bracket_mass_total = bracket_mass * (discrete_inputs["n_front_brackets"] + discrete_inputs["n_rear_brackets"])

        mass_user = float(inputs["spinner_mass_user"][0])
        mass = spin_shell_mass + bracket_mass_total if mass_user == 0.0 else mass_user

        # Compute outputs and assign them to openmdao outputs
        outputs["spinner_diameter"] = sph_spin_diam
        outputs["spinner_mass"] = mass

        # Spinner and hub are assumed to be concentric (spinner wraps hub)
        outputs["spinner_cm"] = inputs["hub_diameter"] / 2.0
        outputs["spinner_cost"] = spin_shell_mass * inputs["composite_cost"] + bracket_mass_total * inputs["metal_cost"]
        outputs["spinner_I"] = np.r_[
            (2.0 / 3.0) * mass * (0.5 * sph_spin_diam) ** 2 * np.ones(3), np.zeros(3)
        ]  # Spherical shell


class PitchSystem(om.ExplicitComponent):
    """
    Semi-empirical model to size the pitch system of the wind turbine rotor.

    Parameters
    ----------
    n_blades : integer
        Number of rotor blades
    blade_mass : float, [kg]
        Total mass of one blade
    rho : float, [kg/m**3]
        Density of the material used for the pitch system
    Xy : float, [Pa]
        Yield strength metal
    pitch_system_scaling_factor : float
        Scaling factor to tune the total mass (0.54 is recommended for modern designs)
    BRFM : float, [N*m]
        Flapwise bending moment at blade root
    hub_diameter : float, [m]
        Outer diameter of the hub

    Returns
    -------
    pitch_mass : float, [kg]
        Total mass of the pitch system
    pitch_cost : float, [USD]
        Cost of the pitch system
    pitch_I : numpy array[6], [kg*m**2]
        Total mass moment of inertia of the pitch system about central point

    """

    def initialize(self):
        self.options.declare("verbosity", default=False)

    def setup(self):
        # Inputs
        self.add_discrete_input("n_blades", val=0)
        self.add_input("blade_mass", val=0.0, units="kg")
        self.add_input("rho", val=0.0, units="kg/m**3")
        self.add_input("Xy", val=0.0, units="Pa")
        self.add_input("pitch_system_scaling_factor", val=0.0)
        self.add_input("BRFM", val=0.0, units="N*m")
        self.add_input("hub_diameter", val=0.0, units="m")
        self.add_input("pitch_system_mass_user", val=0.0, units="kg")
        
        # Outputs
        self.add_output("pitch_mass", val=0.0, units="kg")
        self.add_output("pitch_cost", val=0.0, units="USD")
        self.add_output("pitch_I", val=np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        mass = inputs["pitch_system_scaling_factor"] * (
            0.22 * inputs["blade_mass"] * discrete_inputs["n_blades"]
            + 12.6 * np.abs(inputs["BRFM"]) * inputs["rho"] / inputs["Xy"]
        )
        mass_user = float(inputs["pitch_system_mass_user"][0])
        mass = mass if mass_user == 0.0 else mass_user

        r_hub = 0.5 * inputs["hub_diameter"]
        I = np.r_[mass * r_hub**2 * np.array([1.0, 0.5, 0.5]), np.zeros(3)]

        outputs["pitch_mass"] = mass
        outputs["pitch_cost"] = 0.0
        outputs["pitch_I"] = I

        if self.options["verbosity"]:
            sys.stderr.write(
                "PitchSystem IN : blade mass {:.1f} kg rbmy {:.1f} Nm\n".format(
                    float(inputs["blade_mass"][0]), float(inputs["BRFM"][0])
                )
            )
            sys.stderr.write("PitchSystem OUT: mass {:.1f} kg\n".format(float(mass)))


class Hub_Adder(om.ExplicitComponent):
    """
    Aggregates components for total hub system mass, center of mass, and mass moment of inertia

    Parameters
    ----------
    pitch_mass : float, [kg]
        Total mass of the pitch system
    pitch_cost : float, [kg]
        Cost of the pitch system
    pitch_I : numpy array[6], [kg*m**2]
        Total mass moment of inertia of the pitch system about central point
    hub_mass : float, [kg]
        Total mass of the hub shell, including the flanges
    hub_cost : float, [USD]
        Cost of the hub shell, including flanges
    hub_cm : float, [m]
        Distance between hub/shaft flange and hub center of mass
    hub_I : numpy array[6], [kg*m**2]
        Total mass moment of inertia of the hub about its cm
    spinner_mass : float, [kg]
        Total mass of the spinner
    spinner_cost : float, [kg]
        Cost of the spinner
    spinner_cm : float, [m]
        Radius / Distance between center of mass of the spinner and outer surface
    spinner_I : numpy array[6], [kg*m**2]
        Total mass moment of inertia of the spinner about its cm

    Returns
    -------
    hub_system_mass : float, [kg]
        Total mass of the hub system
    hub_system_cost : float, [USD]
        Cost of the hub system
    hub_system_cm : float, [m]
        Distance between hub/main shaft flange and center of mass of the hub system
    hub_system_I : numpy array[6], [kg*m**2]
        Total mass moment of inertia of the hub system about its center of mass

    """

    def setup(self):
        self.add_input("pitch_mass", val=0.0, units="kg")
        self.add_input("pitch_cost", val=0.0, units="USD")
        self.add_input("pitch_I", val=np.zeros(6), units="kg*m**2")

        self.add_input("hub_mass", val=0.0, units="kg")
        self.add_input("hub_cost", val=0.0, units="USD")
        self.add_input("hub_cm", val=0.0, units="m")
        self.add_input("hub_I", val=np.zeros(6), units="kg*m**2")

        self.add_input("spinner_mass", val=0.0, units="kg")
        self.add_input("spinner_cost", val=0.0, units="kg")
        self.add_input("spinner_cm", val=0.0, units="m")
        self.add_input("spinner_I", val=np.zeros(6), units="kg*m**2")

        self.add_input("hub_system_mass_user", val=0.0, units="kg")
        self.add_input("hub_system_cm_user", val=0.0, units="m")
        self.add_input("hub_system_I_user", val=np.zeros(6), units="kg*m**2")

        self.add_output("hub_system_mass", val=0.0, units="kg")
        self.add_output("hub_system_cost", val=0.0, units="USD")
        self.add_output("hub_system_cm", val=0.0, units="m")
        self.add_output("hub_system_I", val=np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Unpack inputs
        m_pitch = float(inputs["pitch_mass"][0])
        m_hub = float(inputs["hub_mass"][0])
        m_spin = float(inputs["spinner_mass"][0])
        m_user = float(inputs["hub_system_mass_user"][0])
        cm_hub = float(inputs["hub_cm"][0])
        cm_spin = float(inputs["spinner_cm"][0])
        cm_user = inputs["hub_system_cm_user"]
        I_user = inputs["hub_system_cm_user"]

        # Mass and cost totals
        m_total = m_pitch + m_hub + m_spin
        coeff = 1.0 if m_user == 0.0 else m_user / m_total
        m_total *= coeff
        c_total = coeff * inputs["pitch_cost"] + inputs["hub_cost"] + inputs["spinner_cost"]

        # CofM total
        cm_total = ((m_pitch + m_hub) * cm_hub + m_spin * cm_spin) / m_total
        if cm_user != 0.0:
            cm_total = cm_user
        
        # I total
        I_total = util.assembleI(np.zeros(6))
        r = np.array([cm_hub - cm_total, 0.0, 0.0])
        I_total += util.assembleI(inputs["hub_I"]) + m_hub * (np.dot(r, r) * np.eye(3) - np.outer(r, r))
        I_total += util.assembleI(inputs["pitch_I"]) + m_pitch * (np.dot(r, r) * np.eye(3) - np.outer(r, r))
        r = np.array([cm_spin - cm_total, 0.0, 0.0])
        I_total += util.assembleI(inputs["spinner_I"]) + m_spin * (np.dot(r, r) * np.eye(3) - np.outer(r, r))
        I_total *= coeff
        if np.any(I_user):
            I_total = I_user
        
        # Outputs
        outputs["hub_system_mass"] = m_total
        outputs["hub_system_cost"] = c_total
        outputs["hub_system_cm"] = cm_total
        outputs["hub_system_I"] = util.unassembleI(I_total)


class Hub_System(om.Group):
    """
    Openmdao group to model the hub system, which includes pitch system, spinner, and hub
    """

    def initialize(self):
        # self.options.declare('topLevelFlag', default=True)
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]

        self.add_subsystem("maxtorq", FindMaxTorque(), promotes=["*"])
        self.add_subsystem(
            "hub_shell",
            HubShell(gamma=opt["hub_gamma"]),
            promotes=[
                "hub_mass",
                "hub_diameter",
                "hub_cost",
                "hub_cm",
                "hub_I",
                "flange_t2shell_t",
                "flange_OD2hub_D",
                "flange_ID2flange_OD",
                "hub_in2out_circ",
                "blade_root_diameter",
                "hub_stress_concentration",
                "max_torque",
                "constr_hub_diameter",
                "hub_shell_mass_user",
            ],
        )
        self.add_subsystem(
            "spinner",
            Spinner(gamma=opt["spinner_gamma"]),
            promotes=[
                "n_blades",
                "hub_diameter",
                "spinner_mass",
                "spinner_mass_user",
                "spinner_cost",
                "spinner_cm",
                "spinner_I",
                "n_front_brackets",
                "n_rear_brackets",
                "clearance_hub_spinner",
                "spin_hole_incr",
                "blade_root_diameter",
                "spinner_gust_ws",
            ],
        )
        self.add_subsystem(
            "pitch_system",
            PitchSystem(),
            promotes=[
                "n_blades",
                "hub_diameter",
                "pitch_mass",
                "pitch_cost",
                "pitch_I",
                "blade_mass",
                "pitch_system_scaling_factor",
                "pitch_system_mass_user",
            ],
        )
        self.add_subsystem(
            "adder",
            Hub_Adder(),
            promotes=[
                "hub_mass",
                "hub_cost",
                "hub_cm",
                "hub_I",
                "spinner_mass",
                "spinner_cost",
                "spinner_cm",
                "spinner_I",
                "pitch_mass",
                "pitch_cost",
                "pitch_I",
                "hub_system_mass_user",
                "hub_system_cm_user",
                "hub_system_I_user",
                "hub_system_mass",
                "hub_system_cost",
                "hub_system_cm",
                "hub_system_I",
            ],
        )


if __name__ == "__main__":
    hub_prob = om.Problem(model=Hub_System(), reports=False)
    hub_prob.setup()

    hub_prob["n_blades"] = 3
    hub_prob["flange_t2shell_t"] = 4.0
    hub_prob["flange_OD2hub_D"] = 0.5
    hub_prob["flange_ID2flange_OD"] = 0.8
    hub_prob["hub_in2out_circ"] = 1.2
    hub_prob["hub_stress_concentration"] = 2.5
    hub_prob["n_front_brackets"] = 3
    hub_prob["n_rear_brackets"] = 3
    hub_prob["clearance_hub_spinner"] = 0.5
    hub_prob["spin_hole_incr"] = 1.2
    hub_prob["blade_root_diameter"] = 4.5

    hub_prob["pitch_system.blade_mass"] = 17000.0
    hub_prob["pitch_system.BRFM"] = 1.0e6
    hub_prob["pitch_system.scaling_factor"] = 0.54
    hub_prob["pitch_system.rho"] = 7850.0
    hub_prob["pitch_system.Xy"] = 371.0e6

    hub_prob["hub_shell.rho"] = 7200.0
    hub_prob["hub_shell.max_torque"] = 199200777.51  # Value assumed during model development
    hub_prob["hub_shell.Xy"] = 200.0e6
    hub_prob["hub_shell.gamma"] = 2.0
    hub_prob["hub_shell.metal_cost"] = 3.00

    hub_prob["spinner_gust_ws"] = 70
    hub_prob["spinner.gamma"] = 1.5
    hub_prob["spinner.composite_Xt"] = 60.0e6
    # hub_prob['spinner.composite_SF']       = 1.5
    hub_prob["spinner.composite_rho"] = 1600.0
    hub_prob["spinner.Xy"] = 225.0e6
    # hub_prob['spinner.metal_SF']           = 1.5
    hub_prob["spinner.metal_rho"] = 7850.0
    hub_prob["spinner.composite_cost"] = 7.00
    hub_prob["spinner.metal_cost"] = 3.00

    hub_prob.run_model()

    print("Pitch system mass: " + str(hub_prob["pitch_system.pitch_mass"][0]) + " kg")
    print("Hub shell mass: " + str(hub_prob["hub_shell.hub_mass"][0]) + " kg")
    print("Hub shell outer diameter: " + str(hub_prob["hub_shell.hub_diameter"][0]) + " m")
    print("Hub shell cost: " + str(hub_prob["hub_shell.hub_cost"][0]) + " USD")
    print("Distance btw flange and cm of hub shell: " + str(hub_prob["hub_shell.hub_cm"][0]) + " m")
    print("Mass moment of inertia of hub shell: " + str(hub_prob["hub_shell.hub_I"]) + "kg * m2")
    print("Spinner mass: " + str(hub_prob["spinner.spinner_mass"][0]) + " kg")
    print("Spinner outer diameter: " + str(hub_prob["spinner.spinner_diameter"][0]) + " m")
    print("Spinner cost: " + str(hub_prob["spinner.spinner_cost"][0]) + " USD")
    print("Distance btw flange and cm of spinner: " + str(hub_prob["spinner.spinner_cm"][0]) + " m")
    print("Mass moment of inertia of spinner: " + str(hub_prob["spinner.spinner_I"]) + "kg * m2")
    print("Overall hub system mass: " + str(hub_prob["adder.hub_system_mass"][0]) + " kg")
    print("Overall hub system cost: " + str(hub_prob["adder.hub_system_cost"][0]) + " USD")
    print("Distance btw shaft flange and cm of overall hub system: " + str(hub_prob["adder.hub_system_cm"][0]) + " m")
    print("Mass moment of inertia of the overall hub system: " + str(hub_prob["adder.hub_system_I"]) + "kg * m2")
